"""Test picasso functions related to undrifting.

:author: Rafal Kowalewski, 2025
:copyright: Copyright (c) 2025 Jungmann Lab, MPI of Biochemistry
"""

import pytest
from picasso import io, postprocess, aim, lib

# parameters for undrifting
SEGMENTATION = 100


@pytest.fixture(scope="module")
def locs_data():
    """Load localization data once per test module."""
    locs_data = io.load_locs("./tests/data/testdata_locs.hdf5")
    return locs_data


@pytest.fixture(scope="module")
def locs(locs_data):
    """Get locs for testing clusterers."""
    locs = locs_data[0]
    return locs


@pytest.fixture(scope="module")
def info(locs_data):
    """Get info for testing clusterers."""
    info = locs_data[1]
    return info


# undrifting tests
def test_aim(locs, info):
    """Test undrifting by AIM."""
    undrifted_locs, _, drift = aim.aim(locs, info, segmentation=SEGMENTATION)
    assert len(drift) == lib.get_from_metadata(
        info, "Frames"
    ), "Drift length does not match number of frames."
    assert undrifted_locs.shape == locs.shape, "Undrifted locs shape mismatch."


def test_rcc(locs, info):
    """Test undrifting by RCC."""
    drift, undrifted_locs = postprocess.undrift(
        locs, info, segmentation=SEGMENTATION, display=False
    )
    assert len(drift) == lib.get_from_metadata(
        info, "Frames"
    ), "Drift length does not match number of frames."
    assert undrifted_locs.shape == locs.shape, "Undrifted locs shape mismatch."


def test_undrift_from_picked(locs, info):
    """Test undrifting from picked locs."""
    picks = [
        [5.5, 5.5],
        [5.5, 15.5],
        [5.5, 25.5],
        [15.5, 5.5],
        [15.5, 15.5],
        [15.5, 25.5],
        [25.5, 5.5],
        [25.5, 15.5],
        [25.5, 25.5],
    ]
    picked_locs = postprocess.picked_locs(
        locs, info, picks, pick_shape="Circle", pick_size=200 / 130
    )
    drift = postprocess.undrift_from_picked(picked_locs, info)
    assert len(drift) == lib.get_from_metadata(
        info, "Frames"
    ), "Drift length does not match number of frames."



"""
picasso.clusterer
~~~~~~~~~~~~~~~~~

Clusterer for single molecules optimized for DNA-PAINT.
Additionally, contains implementations of DBSCAN and HDBSCAN and other
clustering-related functions.

SMLM clusterer is based on:
* Schlichthaerle, et al. Nature Comm, 2021
  (DOI: 10.1038/s41467-021-22606-1)
* Reinhardt, Masullo, Baudrexel, Steen, et al. Nature, 2023
  (DOI: 10.1038/s41586-023-05925-9)

:authors: Rafal Kowalewski, Susanne Reinhardt,
    Thomas Schlichthaerle, 2020-2025
:copyright: Copyright (c) 2022-2025 Jungmann Lab, MPI of Biochemistry
"""

from __future__ import annotations

from typing import Callable

import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy.spatial import ConvexHull, KDTree, QhullError
from scipy.ndimage import gaussian_filter
from sklearn.cluster import DBSCAN, HDBSCAN

from . import lib, masking


def _frame_analysis(frame: pd.SeriesGroupBy, n_frames: int) -> int:
    """Verify which clusters pass basic frame analysis. Reject clusters
    whose mean frame is outside of the [20, 80] % (max frame) range or
    any 1/20th of measurement's time contains more than 80 % of
    localizations.

    Assumes frame to be a ``pandas.SeriesGroupBy`` object, grouped by
    cluster ids.

    Parameters
    ----------
    frame : pandas.SeriesGroupBy
        Frame number for a given cluster; grouped by cluster ids.
    n_frames : int
        Acquisition time given in frames.

    Returns
    -------
    passed : int
        1 if passed frame analysis, 0 otherwise.
    """
    passed = 1

    # get mean frame
    mean_frame = frame.mean()

    # get maximum number of locs in a 1/20th of acquisition time
    n_locs = len(frame)
    locs_binned = np.histogram(frame, bins=np.linspace(0, n_frames, 21))[0]
    max_locs_bin = locs_binned.max()

    # test if frame analysis passed
    if (
        (mean_frame < 0.2 * n_frames)
        or (mean_frame > 0.8 * n_frames)
        or (max_locs_bin > 0.8 * n_locs)
    ):
        passed = 0

    return passed


def frame_analysis(labels: np.ndarray, frame: np.ndarray) -> np.ndarray:
    """Perform basic frame analysis on clustered localizations. Reject
    clusters whose mean frame is outside of the [20, 80] % (max frame)
    range or any 1/20th of measurement's time contains more than 80 % of
    localizations.

    Uses ``pandas`` for fast calculations using groupby().

    Parameters
    ----------
    labels : np.ndarray
        Cluster labels (-1 means no cluster assigned).
    frame : np.ndarray
        Frame number for each localization.

    Returns
    -------
    labels : np.ndarray
        Cluster labels for each localization (-1 means no cluster
        assigned).
    """
    # group frames by cluster ids
    frame_pd = pd.Series(frame, index=labels)
    frame_grouped = frame_pd.groupby(frame_pd.index)

    # perform frame analysis
    true_cluster = frame_grouped.apply(_frame_analysis, frame.max() + 1)

    # cluster ids that did not pass frame analysis
    discard = true_cluster.index[true_cluster == 0]
    # change labels of these clusters to -1
    labels[np.isin(labels, discard)] = -1

    return labels


def _cluster(
    X: np.ndarray,
    radius: float,
    min_locs: int,
    frame: np.ndarray | None = None,
) -> np.ndarray:
    """Cluster points given by X with a given clustering radius and
    minimum number of localizations within that radius using KDTree.

    The general workflow is as follows:
    1. Build a KDTree from the points in X.
    2. For each point, find its neighbors within the given radius.
    3. Identify local maxima, i.e., points with the most neighbors
       within their neighborhood.
    4. Assign cluster labels to all points. If two local maxima are
       within the radius from each other, combine such clusters.

    Based on the algorithm published by Schichthaerle et al.
    Nature Comm, 2021 (10.1038/s41467-021-22606-1) and first implemented
    in Reinhardt, Masullo, Baudrexel, Steen, et al. Nature, 2023
    (10.1038/s41586-023-05925-9).

    Parameters
    ----------
    X : np.ndarray
        Array of points of shape (n_points, n_dim) to be clustered.
    radius : float
        Clustering radius.
    min_locs : int
        Minimum number of localizations in a cluster.
    frame : np.ndarray, optional
        Frame number of each localization. If None, no frame analysis
        is performed.

    Returns
    -------
    labels : np.ndarray
        Cluster labels for each localization (-1 means no cluster
        assigned).
    """
    # build kdtree (use cKDTree in case user did not update scipy)
    tree = KDTree(X)

    # find neighbors for each point within radius
    neighbors = tree.query_ball_tree(tree, radius)

    # find local maxima, i.e., points with the most neighbors within
    # their neighborhood
    lm = np.zeros(X.shape[0], dtype=np.int8)
    for i in range(len(lm)):
        idx = neighbors[i]  # indeces of points that are neighbors of i
        n = len(idx)  # number of neighbors of i
        if n > min_locs:  # note that i is included in its neighbors
            # if i has the most neighbors in its neighborhood
            if n == max([len(neighbors[_]) for _ in idx]):
                lm[i] = 1

    # assign cluster labels to all points (-1 means no cluster)
    # if two local maxima are within radius from each other, combine
    # such clusters
    labels = -1 * np.ones(X.shape[0], dtype=np.int32)  # cluster labels
    lm_idx = np.where(lm == 1)[0]  # indeces of local maxima

    for count, i in enumerate(lm_idx):  # for each local maximum
        label = labels[i]
        if label == -1:  # if lm not assigned yet
            labels[neighbors[i]] = count
        else:
            # indeces of locs that were not assigned to any cluster
            idx = [
                neighbors[i][_]
                for _ in np.where(labels[neighbors[i]] == -1)[0]
            ]
            if len(idx):  # if such a loc exists, assign it to a cluster
                labels[idx] = label

    # check for number of locs per cluster to be above min_locs
    values, counts = np.unique(labels, return_counts=True)
    # labels to discard if has fewer locs than min_locs
    to_discard = values[counts < min_locs]
    # substitute this with -1
    labels[np.isin(labels, to_discard)] = -1

    if frame is not None:
        # must convert frames to an array, do not change!
        labels = frame_analysis(labels, frame.to_numpy())

    return labels


def cluster_2D(
    locs: pd.DataFrame,
    radius: float,
    min_locs: int,
    fa: bool,
) -> np.ndarray:
    """Prepare 2D input to be used by ``_cluster``.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations to be clustered.
    radius : float
        Clustering radius.
    min_locs : int
        Minimum number of localizations in a cluster.
    fa : bool
        True, if basic frame analysis is to be performed.

    Returns
    -------
    labels : np.ndarray
        Cluster labels for each localization (-1 means no cluster
        assigned).
    """
    X = locs[["x", "y"]].to_numpy()

    if not fa:
        frame = None
    else:
        frame = locs["frame"]

    labels = _cluster(X, radius, min_locs, frame)

    return labels


def cluster_3D(
    locs: pd.DataFrame,
    radius_xy: float,
    radius_z: float,
    min_locs: int,
    fa: bool,
):
    """Prepare 3D input to be used by ``_cluster``.

    Scales z coordinates by radius_xy / radius_z

    Parameters
    ----------
    x, y, z : np.ndarray
        x, y, and z coordinates to be clustered.
    frame : np.ndarray
        Frame number for each localization.
    radius_xy : float
        Clustering radius in x and y directions.
    radius_z : float
        Clustering radius in z direction.
    min_locs : int
        Minimum number of localizations in a cluster.
    fa : bool
        True, if basic frame analysis is to be performed.

    Returns
    -------
    labels : np.ndarray
        Cluster labels for each localization (-1 means no cluster
        assigned).
    """
    radius = radius_xy
    X = locs[["x", "y", "z"]].to_numpy()
    X[:, 2] *= radius_xy / radius_z

    if not fa:
        frame = None
    else:
        frame = locs["frame"]

    labels = _cluster(X, radius, min_locs, frame)

    return labels


def cluster(
    locs: pd.DataFrame,
    radius_xy: float,
    min_locs: int,
    frame_analysis: bool,
    radius_z: float | None = None,
    pixelsize: float | None = None,
) -> pd.DataFrame:
    """Cluster localizations from single molecules (SMLM clusterer).

    The general workflow is as follows:
    1. Build a KDTree from the points in X.
    2. For each point, find its neighbors within the given radius.
    3. Identify local maxima, i.e., points with the most neighbors
       within their neighborhood.
    4. Assign cluster labels to all points. If two local maxima are
       within the radius from each other, combine such clusters.

    Based on the algorithm published by Schichthaerle et al.
    Nature Comm, 2021 (10.1038/s41467-021-22606-1) and first implemented
    in Reinhardt, Masullo, Baudrexel, Steen, et al. Nature, 2023
    (10.1038/s41586-023-05925-9).

    The recommended parameters are ``radius`` of 2*NeNA and ``min_locs``
    of 10. Keep in mind that the parameters may vary between
    applications, so we encourage you to experiment with them when
    needed. Especially ``min_locs`` may vary between datasets.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations to be clustered.
    radius_xy : float
        Clustering radius in xy plane (camera pixels).
    min_locs : int
        Minimum number of localizations in a cluster.
    frame_analysis : bool
        If True, performs basic frame analysis.
    radius_z : float, optional
        Clustering radius in z plane (camera pixels). Only used for
        3D clustering.
    pixelsize : int, optional
        Camera pixel size in nm. Only needed for 3D clustering.

    Returns
    -------
    locs : pd.DataFrame
        Clusterered localizations, with column 'group' added, which
        specifies cluster label for each localization. Noise (label -1)
        is removed.
    """
    locs = locs.copy()
    if "z" in locs.columns:  # 3D
        if pixelsize is None or radius_z is None:
            raise ValueError(
                "Camera pixel size and clustering radius in z must be"
                " specified for 3D clustering."
            )
        locs["z"] /= pixelsize  # convert from nm to px
        labels = cluster_3D(
            locs,
            radius_xy,
            radius_z,
            min_locs,
            frame_analysis,
        )
    else:
        labels = cluster_2D(
            locs,
            radius_xy,
            min_locs,
            frame_analysis,
        )
    locs = extract_valid_labels(locs, labels)
    if "z" in locs.columns:
        locs["z"] *= pixelsize  # convert back to nm
    return locs


def _dbscan(
    X: np.ndarray,
    radius: float,
    min_density: int,
    min_locs: int = 0,
) -> np.ndarray:
    """Find DBSCAN cluster labels, given data points and parameters.

    See Ester, et al. Inkdd, 1996. (Vol. 96, No. 34, pp. 226-231).

    Parameters
    ----------
    X : np.ndarray
        Array of shape (N, D), with N being the number of data points
        and D the number of dimensions.
    radius : float
        DBSCAN search radius, often referred to as "epsilon".
    min_density : int
        Number of points within radius to consider a given point a core
        sample.
    min_locs : int, optional
        Minimum number of localizations in a cluster. Clusters with
        fewer localizations will be removed. Default is 0.

    Returns
    -------
    labels : np.ndarray
        Cluster labels for each point. Shape: (N,). -1 means no cluster
        assigned.
    """
    db = DBSCAN(eps=radius, min_samples=min_density).fit(X)
    labels = db.labels_.astype(np.int32)
    unique_clusters, counts = np.unique(labels, return_counts=True)
    to_discard = unique_clusters[counts < min_locs]
    labels[np.isin(labels, to_discard)] = -1
    return labels


def dbscan(
    locs: pd.DataFrame,
    radius: float,
    min_samples: int,
    min_locs: int = 10,
    pixelsize: int | None = None,
) -> pd.DataFrame:
    """Perform DBSCAN on localizations.

    See Ester, et al. Inkdd, 1996. (Vol. 96, No. 34, pp. 226-231).

    Parameters
    ---------
    locs : pd.DataFrame
        Localizations to be clustered.
    radius : float
        DBSCAN search radius, often referred to as "epsilon". Same units
        as locs.
    min_samples : int
        Number of localizations within radius to consider a given point
        a core sample.
    min_locs : int, optional
        Minimum number of localizations in a cluster. Clusters with
        fewer localizations will be removed. Default is 0.
    pixelsize : int, optional
        Camera pixel size in nm. Only needed for 3D.

    Returns
    -------
    locs : pd.DataFrame
        Clusterered localizations, with column 'group' added, which
        specifies cluster label for each localization. Noise (label -1)
        is removed.
    """
    locs = locs.copy()
    if "z" in locs.columns:
        if pixelsize is None:
            raise ValueError(
                "Camera pixel size must be specified as an integer for 3D"
                " clustering."
            )
        X = locs[["x", "y", "z"]].to_numpy()
        X[:, 2] /= pixelsize
    else:
        X = locs[["x", "y"]].to_numpy()
    labels = _dbscan(X, radius, min_samples, min_locs)
    locs = extract_valid_labels(locs, labels)
    return locs


def _hdbscan(
    X: np.ndarray,
    min_cluster_size: int,
    min_samples: int,
    cluster_eps: float = 0,
) -> np.ndarray:
    """Find HDBSCAN cluster labels, given data points and parameters.

    See Campello, et al. PAKDD, 2013 (DOI: 10.1007/978-3-642-37456-2_14).

    Parameters
    ----------
    X : np.ndarray
        Array of shape (N, D), with N being the number of data points
        and D the number of dimensions.
    min_cluster_size : int
        Minimum number of points in cluster.
    min_samples : int
        Number of points within radius to consider a given point a core
        sample.
    cluster_eps : float, optional
        Distance threshold. Clusters below this value will be merged.

    Returns
    -------
    labels : np.ndarray
        Cluster labels for each point. Shape: (N,). -1 means no cluster
        assigned.
    """
    hdb = HDBSCAN(
        min_samples=min_samples,
        min_cluster_size=min_cluster_size,
        cluster_selection_epsilon=cluster_eps,
    ).fit(X)
    return hdb.labels_.astype(np.int32)


def hdbscan(
    locs: pd.DataFrame,
    min_cluster_size: int,
    min_samples: int,
    pixelsize: int | None = None,
    cluster_eps: float = 0.0,
) -> pd.DataFrame:
    """Perform HDBSCAN on localizations.

    See Campello, et al. PAKDD, 2013 (DOI: 10.1007/978-3-642-37456-2_14).

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations to be clustered.
    min_cluster_size : int
        Minimum number of localizations in cluster.
    min_samples : int
        Number of localizations within radius to consider a given point
        a core sample.
    pixelsize : int, optional
        Camera pixel size in nm. Only needed for 3D.
    cluster_eps : float, optional
        Distance threshold. Clusters below this value will be merged.

    Returns
    -------
    locs : pd.DataFrame
        Clusterered localizations, with column 'group' added, which
        specifies cluster label for each localization. Noise (label -1)
        is removed.
    """
    locs = locs.copy()
    if "z" in locs.columns:
        if pixelsize is None:
            raise ValueError(
                "Camera pixel size (nm) must be specified as an integer for 3D"
                " clustering."
            )
        X = locs[["x", "y", "z"]].to_numpy()
        X[:, 2] /= pixelsize
    else:
        X = locs[["x", "y"]].to_numpy()
    labels = _hdbscan(
        X, min_cluster_size, min_samples, cluster_eps=cluster_eps
    )
    locs = extract_valid_labels(locs, labels)
    return locs


def extract_valid_labels(
    locs: pd.DataFrame,
    labels: np.ndarray,
) -> pd.DataFrame:
    """Extract localizations based on clustering results. Localizations
    that were not clustered are excluded.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations to be filtered.
    labels : np.ndarray
        Array of cluster labels for each localization. -1 means no
        cluster assignment.

    Returns
    -------
    locs : pd.DataFrame
        Localization list with "group" column appended, providing
        cluster label.
    """
    # add cluster id to locs, as "group"
    locs["group"] = labels

    # -1 means no cluster assigned to a loc
    locs = locs[locs["group"] != -1]
    return locs


# def error_sums_wtd(x: float, w: float) -> float:
#     """Find "localization precision" for cluster centers, i.e., weighted
#     standard error of the mean of the localizations in the given
#     cluster.

#     Parameters
#     ----------
#     x : float
#         x or y coordinate of the cluster center.
#     w : float
#         weight (inverse localization precision squared).

#     Returns
#     -------
#     lp : float
#         Weighted standard error of the mean of the cluster center.
#     """
#     lp = (w * (x - (w * x).sum() / w.sum())**2).sum() / w.sum()
#     return lp


def find_cluster_centers(
    locs: pd.DataFrame,
    pixelsize: float | None = None,
) -> pd.DataFrame:
    """Calculate cluster centers.

    Uses ``pandas.groupby`` to quickly run across all cluster ids.

    Parameters
    ----------
    locs : pd.DataFrame
        Clustered localizations (contain group info)
    pixelsize : int, optional
        Camera pixel size (used for finding volume and 3D convex hull).
        Only required for 3D localizations.

    Returns
    -------
    centers : pd.DataFrame
        Cluster centers saved in the format of localizations.
    """
    # group locs by their cluster id (group)
    grouplocs = locs.groupby(locs["group"])

    # get cluster centers
    res = grouplocs.apply(cluster_center, pixelsize, include_groups=False)
    centers_ = res.values

    # convert to DataFrame and save
    frame = np.array([_[0] for _ in centers_])
    std_frame = np.array([_[1] for _ in centers_])
    x = np.array([_[2] for _ in centers_])
    y = np.array([_[3] for _ in centers_])
    std_x = np.array([_[4] for _ in centers_])
    std_y = np.array([_[5] for _ in centers_])
    photons = np.array([_[6] for _ in centers_])
    sx = np.array([_[7] for _ in centers_])
    sy = np.array([_[8] for _ in centers_])
    bg = np.array([_[9] for _ in centers_])
    lpx = np.array([_[10] for _ in centers_])
    lpy = np.array([_[11] for _ in centers_])
    ellipticity = np.array([_[12] for _ in centers_])
    net_gradient = np.array([_[13] for _ in centers_])
    n = np.array([_[14] for _ in centers_])
    n_events = np.array([_[15] for _ in centers_])  # number of locs in cluster

    if "z" in locs.columns:
        z = np.array([_[16] for _ in centers_])
        std_z = np.array([_[17] for _ in centers_])
        volume = np.array([_[18] for _ in centers_])
        convexhull = np.array([_[19] for _ in centers_])
        centers = pd.DataFrame(
            {
                "frame": frame.astype(np.float32),
                "std_frame": std_frame.astype(np.float32),
                "x": x.astype(np.float32),
                "y": y.astype(np.float32),
                "std_x": std_x.astype(np.float32),
                "std_y": std_y.astype(np.float32),
                "z": z.astype(np.float32),
                "photons": photons.astype(np.float32),
                "sx": sx.astype(np.float32),
                "sy": sy.astype(np.float32),
                "bg": bg.astype(np.float32),
                "lpx": lpx.astype(np.float32),
                "lpy": lpy.astype(np.float32),
                "std_z": std_z.astype(np.float32),
                "ellipticity": ellipticity.astype(np.float32),
                "net_gradient": net_gradient.astype(np.float32),
                "n": n.astype(np.uint32),
                "n_events": n_events.astype(np.int32),
                "volume": volume.astype(np.float32),
                "convexhull": convexhull.astype(np.float32),
                "group": res.index.astype(np.int32),  # group id
            }
        )
    else:
        area = np.array([_[16] for _ in centers_])
        convexhull = np.array([_[17] for _ in centers_])
        centers = pd.DataFrame(
            {
                "frame": frame.astype(np.float32),
                "std_frame": std_frame.astype(np.float32),
                "x": x.astype(np.float32),
                "y": y.astype(np.float32),
                "std_x": std_x.astype(np.float32),
                "std_y": std_y.astype(np.float32),
                "photons": photons.astype(np.float32),
                "sx": sx.astype(np.float32),
                "sy": sy.astype(np.float32),
                "bg": bg.astype(np.float32),
                "lpx": lpx.astype(np.float32),
                "lpy": lpy.astype(np.float32),
                "ellipticity": ellipticity.astype(np.float32),
                "net_gradient": net_gradient.astype(np.float32),
                "n": n.astype(np.uint32),
                "n_events": n_events.astype(np.int32),
                "area": area.astype(np.float32),
                "convexhull": convexhull.astype(np.float32),
                "group": res.index.astype(np.int32),  # group id
            }
        )
    if "group_input" in locs.columns:
        group_input = np.array([_[-1] for _ in centers_])
        centers["group_input"] = group_input.astype(np.int32)

    return centers


def cluster_center(
    grouplocs: pd.SeriesGroupBy,
    pixelsize: float | None = None,
    separate_lp: bool = False,
) -> pd.Series:
    """Find cluster centers and their attributes, such as mean number
    of photons per localization, etc.

    Assumes locs to be a ``pandas.SeriesGroupBy`` object, grouped by
    cluster ids.

    Parameters
    ----------
    grouplocs : pandas.SeriesGroupBy
        Localizations grouped by cluster ids.
    pixelsize : int, optional
        Camera pixel size (used for finding volume and 3D convex hull).
        Only required for 3D localizations.
    separate_lp : bool, optional
        If True, localization precision in x and y will be calculated
        separately. Otherwise, the mean of the two is taken.

    Returns
    -------
    results : pd.Series
        Cluster center attributes. For each group, a list of values is
        returned: x, y, (z, optional), etc.
    """
    # mean and std frame
    frame = grouplocs.frame.mean()
    std_frame = grouplocs.frame.std()
    # average x and y, weighted by lpx, lpy
    # x = np.average(grouplocs.x, weights=1/(grouplocs.lpx)**2)
    # y = np.average(grouplocs.y, weights=1/(grouplocs.lpy)**2)
    x = np.mean(grouplocs.x)
    y = np.mean(grouplocs.y)
    std_x = grouplocs.x.std()
    std_y = grouplocs.y.std()
    # mean values
    photons = grouplocs.photons.mean()
    sx = grouplocs.sx.mean()
    sy = grouplocs.sy.mean()
    bg = grouplocs.bg.mean()
    # weighted mean loc precision
    # lpx = np.sqrt(
    #     error_sums_wtd(grouplocs.x, grouplocs.lpx)
    #     / (len(grouplocs) - 1)
    # )
    # lpy = np.sqrt(
    #     error_sums_wtd(grouplocs.y, grouplocs.lpy)
    #     / (len(grouplocs) - 1)
    # )
    lpx = np.std(grouplocs.x) / len(grouplocs) ** 0.5
    lpy = np.std(grouplocs.y) / len(grouplocs) ** 0.5
    if not separate_lp:
        lpx = (lpx + lpy) / 2
        lpy = lpx
    # other attributes
    ellipticity = sx / sy
    net_gradient = grouplocs.net_gradient.mean()
    # n_locs in cluster
    n = len(grouplocs)
    # number of binding events
    split_idx = np.where(np.diff(grouplocs.frame) > 3)[0] + 1  # split locs by
    # consecutive frames
    x_events = np.split(grouplocs.x.to_numpy(), split_idx)
    n_events = len(x_events)  # number of binding events
    if "z" in grouplocs.columns:
        if pixelsize is None:
            raise ValueError(
                "Camera pixel size must be specified as an integer for 3D"
                " cluster centers calculation."
            )
        z = np.average(
            grouplocs.z,
            weights=1 / ((grouplocs.lpx + grouplocs.lpy) ** 2),
        )  # take lpz = 2 * mean(lpx, lpy)
        std_z = grouplocs.z.std()
        # lpz = std_z
        volume = (
            np.power((std_x + std_y + std_z / pixelsize) / 3 * 2, 3) * 4.18879
        )
        try:
            X = np.stack(
                (grouplocs.x, grouplocs.y, grouplocs.z / pixelsize),
                axis=0,
            ).T
            hull = ConvexHull(X)
            convexhull = hull.volume
        except QhullError:
            convexhull = 0
        result = [
            frame,
            std_frame,
            x,
            y,
            std_x,
            std_y,
            photons,
            sx,
            sy,
            bg,
            lpx,
            lpy,
            ellipticity,
            net_gradient,
            n,
            n_events,
            z,
            std_z,
            # lpz,
            volume,
            convexhull,
        ]
    else:
        area = np.power(std_x + std_y, 2) * np.pi
        try:
            X = np.stack((grouplocs.x, grouplocs.y), axis=0).T
            hull = ConvexHull(X)
            convexhull = hull.volume
        except QhullError:
            convexhull = 0
        result = [
            frame,
            std_frame,
            x,
            y,
            std_x,
            std_y,
            photons,
            sx,
            sy,
            bg,
            lpx,
            lpy,
            ellipticity,
            net_gradient,
            n,
            n_events,
            area,
            convexhull,
        ]

    if "group_input" in grouplocs.columns:
        # assumes only one group input!
        result.append(np.unique(grouplocs.group_input)[0])
    return result


def _cluster_area(X: np.ndarray, lp: float) -> float:
    """Calculate cluster area (2D) or volume (3D). Uses Otsu
    thresholding of the images of the clusters to find areas/volumes.

    Parameters
    ----------
    X : np.ndarray
        Array of points of shape (n_points, n_dim).
    lp : float
        Median localization precision in x and y of the dataset. Used to
        define the pixel size for image rendering

    Returns
    -------
    area : float
        Cluster area (2D) or volume (3D) in units of LP.
    """
    # get image
    bin_size = lp / 2  # pixel size for rendering
    if X.shape[1] == 3:  # 3D
        bin_size_z = bin_size * 2.5  # just a rough estimate
        edges = [
            np.arange(X[:, 0].min(), X[:, 0].max() + bin_size, bin_size),
            np.arange(X[:, 1].min(), X[:, 1].max() + bin_size, bin_size),
            np.arange(X[:, 2].min(), X[:, 2].max() + bin_size_z, bin_size_z),
        ]
    else:  # 2D
        edges = [
            np.arange(X[:, 0].min(), X[:, 0].max() + bin_size, bin_size),
            np.arange(X[:, 1].min(), X[:, 1].max() + bin_size, bin_size),
        ]
    image = np.histogramdd(X, bins=edges)[0]
    image = gaussian_filter(image, sigma=2)  # smooth image with sigma = LP
    # threshold the image and calculate area/volume
    thresh = masking.threshold_otsu(image.reshape(-1))
    if X.shape[1] == 3:  # 3D
        area = np.sum(image >= thresh) / (
            16 / 5
        )  # volume in LP^3, bins are 0.5 LP in xy and 1.25 LP in z
    else:  # 2D
        area = np.sum(image >= thresh) / 4  # area in LP^2, bins are 0.5 LP
    return area


def cluster_areas(
    locs: pd.DataFrame,
    info: list[dict],
    progress: Callable[[int], None] | None = None,
) -> pd.DataFrame:
    """Calculate cluster areas (2D) or volumes (3D).

    Uses Otsu thresholding of the images of the clusters to find areas/
    volumes.

    Parameters
    ----------
    locs : pd.DataFrame
        Clustered localizations (contain group info).
    info : list of dict
        Localization metadata, see `picasso.io.load_locs`.
    progress : picasso.lib.ProgressDialog, optional
        Progress dialog. If None, progress is displayed with into the
        console. Default is None.

    Returns
    -------
    areas : pd.DataFrame
        Cluster areas/volumes for each cluster.
    """
    assert (
        "group" in locs.columns
    ), "Localizations must contain 'group' column."

    # get pixel size from info
    pixelsize = lib.get_from_metadata(info, "Pixelsize", raise_error=True)

    groups = np.unique(locs["group"])
    area_key = "Area (LP^2)" if "z" not in locs.columns else "Volume (LP^3)"
    areas = {
        "group": groups.astype(np.int32),
        area_key: np.zeros(len(groups), dtype=np.float32),
    }
    if progress is None:
        iterator = tqdm(range(len(groups)), desc="Calculating cluster areas")
    else:
        iterator = range(len(groups))

    lp = np.median(locs[["lpx", "lpy"]].mean(axis=1))
    for idx in iterator:
        group_id = groups[idx]
        grouplocs = locs[locs["group"] == group_id]
        if not len(grouplocs):
            continue
        if "z" in grouplocs.columns:
            X = grouplocs[["x", "y", "z"]].to_numpy()
            X[:, 2] /= pixelsize  # convert z to pixels
        else:
            X = grouplocs[["x", "y"]].to_numpy()
        areas[area_key][idx] = _cluster_area(X, lp)
        if progress is not None:
            progress(idx + 1)
    return pd.DataFrame(areas)


def test_subclustering(
    mols: pd.DataFrame,
    info: list[dict],
    clustering_dist: float = 25,
    sparse_dist: float = 80,
) -> tuple[np.ndarray, np.ndarray]:
    """Extract number of events from molecular maps based on their
    numbers of binding events assigned.

    The reasoning is that 'subclustered' molecules will tend to have
    fewer binding events assigned to them since multiple molecules
    are assigned to a real single molecule. Thus, we can compare the
    distribution of the number of events for the populations of
    molecules that have close neighbors (clustered) and those that
    do not (sparse). If subclustering occurs, the clustered molecules
    should have a lower number of events on average.

    DOI: *DOI will be added once available*

    Parameters
    ----------
    mols : pd.DataFrame
        List of molecules, must contain n_events.
    info : list of dict
        Molecules metadata.
    clustering_dist : float, optional
        Maximum distance between molecules (nm) to consider them as
        clustered. Default is 25.
    sparse_dist : float, optional
        Minimum distance between molecules (nm) to consider them as sparse.
        Default is 80.

    Returns
    -------
    clustered_nevents : np.ndarray
        Number of events for clustered molecules.
    sparse_nevents : np.ndarray
        Number of events for sparse molecules.
    """
    assert (
        "n_events" in mols.columns
    ), "The input molecules must have n_events attribute."
    assert sparse_dist > clustering_dist, (
        "The sparse distance must be larger than the clustering " "distance."
    )
    pixelsize = lib.get_from_metadata(info, "Pixelsize")
    if pixelsize is None:
        raise ValueError("Pixelsize not found in metadata.")

    # get 1st nearest neighbor distances
    if "z" in mols.columns:
        coords = mols[["x", "y", "z"]].to_numpy()
        coords[:, 2] /= pixelsize
    else:
        coords = mols[["x", "y"]].to_numpy()
    tree = KDTree(coords)
    distances, indices = tree.query(coords, k=2)
    nnd1 = distances[:, 1]
    idx1 = indices[:, 1]

    # split molecules into clustered and monomeric
    close_nnd_idx = np.where(nnd1 < clustering_dist / pixelsize)[0]
    far_nnd_idx = np.where(nnd1 >= sparse_dist / pixelsize)[0]
    close_mols = mols.iloc[close_nnd_idx]
    far_mols = mols.iloc[far_nnd_idx]
    clustered_nevents = close_mols["n_events"].to_numpy()
    sparse_nevents = far_mols["n_events"].to_numpy()
    return clustered_nevents, sparse_nevents



"""
picasso.nanotron
~~~~~~~~~~~~~~~~

Deep learning library for classification of picked localizations.

:author: Alexander Auer, Maximilian Strauss 2020
:copyright: Copyright (c) 2020 Jungmann Lab, MPI of Biochemistry
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from tqdm import tqdm as tqdm
from scipy import ndimage
from sklearn.neural_network import MLPClassifier

from . import render, lib


def prepare_img(
    img: np.ndarray,
    img_shape: int,
    alpha: float = 1,
    bg: float = 0,
) -> np.ndarray:
    """Prepare image for classification.

    Parameters
    ----------
    img : np.ndarray
        Input image to be prepared.
    img_shape : int
        Shape of the image (assumed to be square).
    alpha : float, optional
        Scaling factor for the image, by default 1.
    bg : float, optional
        Background value to be subtracted, by default 0.

    Returns
    -------
    img : np.ndarray
        Prepared image.
    """
    img = alpha * img - bg
    img = img.astype("float")
    img = img / img.max()
    img = img.clip(min=0)
    img = img.reshape(img_shape**2)

    return img


def rotate_img(img: np.ndarray, angle: float) -> np.ndarray:
    """Rotate image by a given angle.

    Parameters
    ----------
    img : np.ndarray
        Input image to be rotated.
    angle : float
        Angle in degrees by which to rotate the image.

    Returns
    -------
    rot_img : np.ndarray
        Rotated image.
    """
    rot_img = ndimage.rotate(img, angle, reshape=False)

    return rot_img


def roi_to_img(
    locs: pd.DataFrame,
    pick: int,
    radius: float,
    oversampling: float,
    picks: tuple[float, float] | None = None,
) -> np.ndarray:
    """Convert a region of interest (ROI) defined by localizations to an
    image.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations from which to create the image.
    pick : int
        The group number of the localizations to be picked.
    radius : float
        Radius around the mean position of the localizations to define
        the ROI.
    oversampling : float
        Number of super-resolution pixels per camera pixel.
    picks : tuple[float, float] | None, optional
        Specific coordinates (x, y) to isolate localizations from, by
        default None. If None, uses the mean position of the localizations
        in the specified group.

    Returns
    -------
    pick_img : np.ndarray
        Image of the picked localizations.
    """
    # Isolate locs from pick
    pick_locs = []
    if picks is None:
        pick_locs = locs[locs["group"] == pick]
    else:
        x, y = picks
        pick_locs = lib.locs_at(x, y, locs, radius)
        pick_locs.sort_values(by="frame", kind="quicksort", inplace=True)
    # dirty method to avoid floating point errors with render
    radius -= 0.001

    x_mean = np.mean(pick_locs["x"])
    y_mean = np.mean(pick_locs["y"])

    x_min = x_mean - radius
    x_max = x_mean + radius
    y_min = y_mean - radius
    y_max = y_mean + radius

    viewport = (y_min, x_min), (y_max, x_max)

    # for debugging
    if False:
        print("mean x: {}".format(np.mean(pick_locs.x)))
        print("length x: {}".format(x_max - x_min))
        print("mean y: {}".format(np.mean(pick_locs.y)))
        print("length y: {}".format(y_max - y_min))
        print("radius: {}".format(radius))
        print("viewport: {}".format(viewport))

    # Render locs with Picasso render function
    try:
        len_x, pick_img = render.render(
            pick_locs,
            viewport=viewport,
            oversampling=oversampling,
            blur_method="smooth",
        )
    except Exception:
        pass
    return pick_img


def prepare_data(
    locs: pd.DataFrame,
    label: int,
    pick_radius: float,
    oversampling: float,
    alpha: float = 10,
    bg: float = 1,
    export: bool = False,
) -> tuple[list[np.ndarray], list[int]]:
    """Prepare data for classification by extracting images of
    localizations.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations from which to create the images.
    label : int
        Label for the data, typically the group number.
    pick_radius : float
        Radius around the mean position of the localizations to define
        the ROI.
    oversampling : float
        Number of super-resolution pixels per camera pixel.
    alpha : float, optional
        Scaling factor for the image, by default 10.
    bg : float, optional
        Background value to be subtracted, by default 1.
    export : bool, optional
        If True, saves the images to the './img/' directory, by default
        False.

    Returns
    -------
    data : list[np.ndarray]
        List of prepared images of the localizations.
    labels : list[int]
        List of labels corresponding to the images.
    """
    img_shape = int(2 * pick_radius * oversampling)
    data = []
    labels = []

    for pick in tqdm(range(locs["group"].max()), desc="Prepare " + str(label)):

        pick_img = roi_to_img(
            locs, pick, radius=pick_radius, oversampling=oversampling
        )

        if export is True and pick < 10:
            filename = "label" + str(label) + "-" + str(pick)
            plt.imsave(
                "./img/" + filename + ".png",
                (alpha * pick_img - bg),
                cmap="Greys",
                vmax=10,
            )

        pick_img = prepare_img(
            pick_img,
            img_shape=img_shape,
            alpha=alpha,
            bg=bg,
        )

        data.append(pick_img)
        labels.append(label)

    return data, label


def predict_structure(
    mlp: MLPClassifier,
    locs: pd.DataFrame,
    pick: int,
    pick_radius: float,
    oversampling: float,
    picks: tuple[float, float] | None = None,
) -> tuple[int, np.ndarray]:
    """Predict the structure of localizations using a trained MLP
    classifier.

    Parameters
    ----------
    mlp : MLPClassifier
        Trained MLP classifier.
    locs : pd.DataFrame
        Localizations to predict.
    pick : int
        Index of the localizations to predict.
    pick_radius : float
        Radius around the localizations to consider.
    oversampling : float
        Oversampling factor for the image.
    picks : tuple[float, float] | None, optional
        Coordinates of the region of interest (ROI) to predict.

    Returns
    -------
    pred : int
        Predicted label for the image.
    pred_proba : np.ndarray
        Predicted probabilities for each class.
    """
    img_shape = int(2 * pick_radius * oversampling)
    img = roi_to_img(
        locs,
        pick=pick,
        radius=pick_radius,
        oversampling=oversampling,
        picks=picks,
    )
    img = prepare_img(img, img_shape=img_shape, alpha=10, bg=1)
    img = img.reshape(1, img_shape**2)

    pred = mlp.predict(img)
    pred_proba = mlp.predict_proba(img)

    return pred, pred_proba



"""
picasso.zfit
~~~~~~~~~~~~

Fitting z coordinates using astigmatism.

:author: Joerg Schnitzbauer, 2016
:copyright: Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry
"""

from __future__ import annotations

import multiprocessing
from concurrent import futures
from concurrent.futures import ProcessPoolExecutor
from typing import Literal

import numba
import yaml
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from tqdm import tqdm
from scipy.optimize import minimize_scalar

from . import lib, gausslq, gaussmle


plt.style.use("ggplot")


def nan_index(y: np.ndarray) -> tuple[np.ndarray, callable]:
    """Find indices of NaN values in an array."""
    return np.isnan(y), lambda z: z.nonzero()[0]


def interpolate_nan(data: np.ndarray) -> np.ndarray:
    """Linear interpolattion of NaN values in an array ``data``."""
    nans, x = nan_index(data)
    data[nans] = np.interp(x(nans), x(~nans), data[~nans])
    return data


def calibrate_z(
    locs: pd.DataFrame,
    info: list[dict],
    d: float,
    magnification_factor: float,
    path: str | None = None,
) -> dict:
    """Given localizations of a calibration sample (e.g., gold beads at
    different z positions), calibrate the z-axis by fitting a polynomial
    to the mean spot width/height of each frame. See Huang et al.
    Science, 2008. DOI: 10.1126/science.1153529.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations of a calibration sample.
    info : list of dicts
        Information about the calibration sample, including the number
        of frames.
    d : float
        Step size in nm, i.e., the distance between the z positions of
        the calibration sample.
    magnification_factor : float
        Magnification factor of the microscope, i.e., the ratio between
        the actual z position of the calibration sample and the
        estimated z position from the localization data.
    path : str, optional
        Path to save the calibration data as a YAML file. If None, the
        calibration data will not be saved. Default is None.

    Returns
    -------
    calibration : dict
        Dictionary containing the calibration coefficients (i.e.,
        polynomial coefficients), number of frames, step size, and
        magnification factor.
    """
    n_frames = info[0]["Frames"]
    range = (n_frames - 1) * d
    frame_range = np.arange(n_frames)
    z_range = -(frame_range * d - range / 2)  # negative so that the
    # first frames of a bottom-to-up scan are positive z coordinates.

    mean_sx = np.array(
        [locs["sx"][locs["frame"] == _].mean() for _ in frame_range]
    )
    mean_sy = np.array(
        [locs["sy"][locs["frame"] == _].mean() for _ in frame_range]
    )
    var_sx = np.array(
        [locs["sx"][locs["frame"] == _].var() for _ in frame_range]
    )
    var_sy = np.array(
        [locs["sy"][locs["frame"] == _].var() for _ in frame_range]
    )

    keep_x = (locs["sx"] - mean_sx[locs["frame"]]) ** 2 < var_sx[locs["frame"]]
    keep_y = (locs["sy"] - mean_sy[locs["frame"]]) ** 2 < var_sy[locs["frame"]]
    keep = keep_x & keep_y
    locs = locs[keep]

    # Fits calibration curve to the mean of each frame
    mean_sx = np.array(
        [locs["sx"][locs["frame"] == _].mean() for _ in frame_range]
    )
    mean_sy = np.array(
        [locs["sy"][locs["frame"] == _].mean() for _ in frame_range]
    )

    # Fix nan
    mean_sx = interpolate_nan(mean_sx)
    mean_sy = interpolate_nan(mean_sy)

    cx = np.polyfit(z_range, mean_sx, 6, full=False)
    cy = np.polyfit(z_range, mean_sy, 6, full=False)

    # make sure that the calibration curves cross at z = 0
    z = np.linspace(z_range[0], z_range[-1], 10000)
    spot_width = np.poly1d(cx)
    spot_height = np.poly1d(cy)
    z_range -= z[np.argmin(np.abs(spot_width(z) - spot_height(z)))]
    cx = np.polyfit(z_range, mean_sx, 6, full=False)
    cy = np.polyfit(z_range, mean_sy, 6, full=False)

    calibration = {
        "X Coefficients": [float(_) for _ in cx],
        "Y Coefficients": [float(_) for _ in cy],
        "Number of frames": int(n_frames),
        "Step size in nm": float(d),
        "Magnification factor": float(magnification_factor),
    }
    if path is not None:
        with open(path, "w") as f:
            yaml.dump(calibration, f, default_flow_style=False)

    # pixelsize does not matter here anyway
    locs = fit_z(locs, info, calibration, magnification_factor, pixelsize=130)
    locs["z"] /= magnification_factor

    plt.figure(figsize=(18, 10))

    plt.subplot(231)
    plt.plot(z_range, mean_sx, ".-", label="x")
    plt.plot(z_range, mean_sy, ".-", label="y")
    plt.plot(z_range, np.polyval(cx, z_range), "0.3", lw=1.5, label="x fit")
    plt.plot(z_range, np.polyval(cy, z_range), "0.3", lw=1.5, label="y fit")
    plt.xlabel("Stage position")
    plt.ylabel("Mean spot width/height")
    plt.xlim(z_range.min(), z_range.max())
    plt.legend(loc="best")

    ax = plt.subplot(232)
    plt.scatter(locs["sx"], locs["sy"], c="k", lw=0, alpha=0.1)
    plt.plot(
        np.polyval(cx, z_range),
        np.polyval(cy, z_range),
        lw=1.5,
        label="calibration from fit of mean width/height",
    )
    plt.plot()
    ax.set_aspect("equal")
    plt.xlabel("Spot width")
    plt.ylabel("Spot height")
    plt.legend(loc="best")

    plt.subplot(233)
    plt.plot(locs["z"], locs["sx"], ".", label="x", alpha=0.2)
    plt.plot(locs["z"], locs["sy"], ".", label="y", alpha=0.2)
    plt.plot(
        z_range,
        np.polyval(cx, z_range),
        "0.3",
        lw=1.5,
        label="calibration",
    )
    plt.plot(z_range, np.polyval(cy, z_range), "0.3", lw=1.5)
    plt.xlim(z_range.min(), z_range.max())
    plt.xlabel("Estimated z")
    plt.ylabel("Spot width/height")
    plt.legend(loc="best")

    ax = plt.subplot(234)
    plt.plot(z_range[locs["frame"]], locs["z"], ".k", alpha=0.1)
    plt.plot(
        [z_range.min(), z_range.max()],
        [z_range.min(), z_range.max()],
        lw=1.5,
        label="identity",
    )
    plt.xlim(z_range.min(), z_range.max())
    plt.ylim(z_range.min(), z_range.max())
    ax.set_aspect("equal")
    plt.xlabel("Stage position")
    plt.ylabel("Estimated z")
    plt.legend(loc="best")

    ax = plt.subplot(235)
    deviation = locs["z"] - z_range[locs["frame"]]
    bins = lib.calculate_optimal_bins(deviation, max_n_bins=1000)
    plt.hist(deviation, bins)
    plt.xlabel("Deviation to true position")
    plt.ylabel("Occurence")

    ax = plt.subplot(236)
    square_deviation = deviation**2
    mean_square_deviation_frame = [
        np.mean(square_deviation[locs["frame"] == _]) for _ in frame_range
    ]
    rmsd_frame = np.sqrt(mean_square_deviation_frame)
    plt.plot(z_range, rmsd_frame, ".-", color="0.3")
    plt.xlim(z_range.min(), z_range.max())
    plt.gca().set_ylim(bottom=0)
    plt.xlabel("Stage position")
    plt.ylabel("Mean z precision")

    plt.tight_layout(pad=2)

    if path is not None:
        dirname = path[0:-5]
        plt.savefig(dirname + ".png", format="png", dpi=300)

    plt.show()
    return calibration


@numba.jit(nopython=True, nogil=True)
def _fit_z_target(
    z: float,
    sx: float,
    sy: float,
    cx: np.ndarray,
    cy: np.ndarray,
) -> float:
    """Target function that's to be minimized for fitting the z
    coordinates given the single-emitter image width and height as well
    as the calibration curve coefficients. It calculates the difference
    between the square root of the spot width/height and the polynomial
    fit of the z-axis calibration curve. Based on Huang et al. Science,
    2008. DOI: 10.1126/science.1153529."""
    z2 = z * z
    z3 = z * z2
    z4 = z * z3
    z5 = z * z4
    z6 = z * z5
    wx = (
        cx[0] * z6
        + cx[1] * z5
        + cx[2] * z4
        + cx[3] * z3
        + cx[4] * z2
        + cx[5] * z
        + cx[6]
    )
    wy = (
        cy[0] * z6
        + cy[1] * z5
        + cy[2] * z4
        + cy[3] * z3
        + cy[4] * z2
        + cy[5] * z
        + cy[6]
    )
    return (sx**0.5 - wx**0.5) ** 2 + (sy**0.5 - wy**0.5) ** 2


def fit_z(
    locs: pd.DataFrame,
    info: list[dict],
    calibration: dict,
    magnification_factor: float,
    pixelsize: float,
    fitting_method: Literal["gausslq", "gaussmle"] = "gausslq",
    filter: int = 2,
) -> pd.DataFrame:
    """Fit z coordinates to the localizations based on the calibration
    curve coefficients and the single-emitter image width and height.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations to fit the z-axis calibration curve to.
    info : list of dicts
        Information about the localizations, including the number of
        frames.
    calibration : dict
        Calibration data containing the polynomial coefficients for
        the x and y axes, number of frames, step size, and magnification
        factor.
    magnification_factor : float
        Magnification factor of the microscope, i.e., the ratio between
        the actual z position of the calibration sample and the
        estimated z position from the localization data.
    pixelsize : float
        Camera pixel size in nm.
    fitting_method : {"gausslq", "gaussmle"}, optional
        Fitting method used to obtain 2D localization parameters (x, y,
        sx, sy). Default is "gausslq".
    filter : int, optional
        Filter for the z fits. If set to 0, no filtering is applied.
        If set to 2, the z fits are filtered based on the root mean
        square deviation (RMSD) of the z calibration. Default is 2.

    Returns
    -------
    locs : pd.DataFrame
        Localizations with the fitted z coordinates, their residuals
        (d_zcalib) and axial localization precision appended.
    """
    locs = locs.copy()
    cx = np.array(calibration["X Coefficients"])
    cy = np.array(calibration["Y Coefficients"])
    # in multiprocessing, pandas Series causes issues!
    sx = locs["sx"].to_numpy()
    sy = locs["sy"].to_numpy()
    z = np.zeros_like(locs["x"])
    square_d_zcalib = np.zeros_like(z)
    for i in range(len(z)):
        # set bounds to avoid potential gaps in the calibration curve,
        # credits to Loek Andriessen
        result = minimize_scalar(
            _fit_z_target,
            bounds=[-1000, 1000],
            args=(sx[i], sy[i], cx, cy),
        )
        z[i] = result.x
        square_d_zcalib[i] = result.fun
    locs["z"] = z * magnification_factor
    locs["d_zcalib"] = np.sqrt(square_d_zcalib)
    lpz = _axial_localization_precision_astig(
        locs,
        cx,
        cy,
        magnification_factor,
        pixelsize,
        fitting_method,
    )
    locs["lpz"] = lpz
    locs = lib.ensure_sanity(locs, info)
    return filter_z_fits(locs, filter)


def fit_z_parallel(
    locs: pd.DataFrame,
    info: list[dict],
    calibration: dict,
    magnification_factor: float,
    pixelsize: float,
    fitting_method: Literal["gausslq", "gaussmle"] = "gausslq",
    filter: int = 2,
    asynch: bool = False,
) -> pd.DataFrame | list[futures.Future]:
    """Fit z coordinates to the localizations based on the calibration
    curve coefficients and the single-emitter image width and height,
    optionally using multiprocessing.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations to fit the z-axis calibration curve to.
    info : list of dicts
        Information about the localizations, including the number of
        frames.
    calibration : dict
        Calibration data containing the polynomial coefficients for
        the x and y axes, number of frames, step size, and magnification
        factor.
    magnification_factor : float
        Magnification factor of the microscope, i.e., the ratio between
        the actual z position of the calibration sample and the
        estimated z position from the localization data.
    pixelsize : float
        Camera pixel size in nm.
    fitting_method : {"gausslq", "gaussmle"}, optional
        Fitting method used to obtain 2D localization parameters (x, y,
        sx, sy). Default is "gausslq".
    filter : int, optional
        Filter for the z fits. If set to 0, no filtering is applied.
        If set to 2, the z fits are filtered based on the root mean
        square deviation (RMSD) of the z calibration. Default is 2.
    asynch : bool, optional
        If True, use multiprocessing. Then, a list of futures that can
        be used to retrieve the results asynchronously is returned. If
        False, the function waits for all tasks to complete and returns
        the combined results. Default is False.

    Returns
    -------
    locs : pd.DataFrame or list of futures.Future
        If `asynch` is False, returns a DataFrame of localizations with
        the fitted z coordinates and their residuals (d_zcalib).
        If `asynch` is True, returns a list of futures that can be
        used to retrieve the results asynchronously.
    """
    n_workers = min(
        60, max(1, int(0.75 * multiprocessing.cpu_count()))
    )  # Python crashes when using >64 cores
    n_locs = len(locs)
    n_tasks = 100 * n_workers
    spots_per_task = [
        (
            int(n_locs / n_tasks + 1)
            if _ < n_locs % n_tasks
            else int(n_locs / n_tasks)
        )
        for _ in range(n_tasks)
    ]
    start_indices = np.cumsum([0] + spots_per_task[:-1])
    fs = []
    executor = ProcessPoolExecutor(n_workers)
    for i, n_locs_task in zip(start_indices, spots_per_task):
        fs.append(
            executor.submit(
                fit_z,
                locs[i : i + n_locs_task],
                info,
                calibration,
                magnification_factor,
                pixelsize,
                fitting_method=fitting_method,
                filter=0,
            )
        )
    if asynch:
        return fs
    with tqdm(total=n_tasks, unit="task") as progress_bar:
        for f in futures.as_completed(fs):
            progress_bar.update()
    return locs_from_futures(fs, filter=filter)


def locs_from_futures(
    futures: list[futures.Future], filter: int = 2
) -> pd.DataFrame:
    """Combine the results from a list of futures (i.e.,
    multiprocessing results) into a single DataFrame of localizations
    with fitted z coordinates and their residuals (d_zcalib).

    Parameters
    ----------
    futures : list of futures.Future
        List of futures that contain the results of the z fits.
    filter : int, optional
        Filter for the z fits. If set to 0, no filtering is applied.
        If set to 2, the z fits are filtered based on the root mean
        square deviation (RMSD) of the z calibration. Default is 2.

    Returns
    -------
    locs : pd.DataFrame
        DataFrame of localizations with the fitted z coordinates and
        their residuals (d_zcalib).
    """
    locs = [_.result() for _ in futures]
    locs = pd.concat(locs, ignore_index=True)
    return filter_z_fits(locs, filter)


def filter_z_fits(locs: pd.DataFrame, range: int) -> pd.DataFrame:
    """Filter the z fits based on the root mean square deviation (RMSD)
    of the z calibration (d_zcalib residual). If `range` is set to 0, no
    filtering is applied. If `range` is greater than 0, the
    localizations with a RMSD greater than `range` are removed.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations with fitted z coordinates and their residuals
        (d_zcalib).
    range : int
        Range for filtering the z fits. If set to 0, no filtering is
        applied. If set to a positive value, localizations with a
        RMSD greater than `range` times the RMSD of the z calibration
        are removed.

    Returns
    -------
    locs : pd.DataFrame
        DataFrame of localizations with the fitted z coordinates and
        their residuals (d_zcalib) after filtering.
    """
    if "d_zcalib" not in locs.columns:
        return locs
    if range > 0:
        rmsd = np.sqrt(np.nanmean(locs["d_zcalib"] ** 2))
        locs = locs[locs["d_zcalib"] <= range * rmsd]
    return locs


def axial_localization_precision(
    locs: np.recarray,
    info: list[dict],
    calibration: dict,
    fitting_method: Literal["gausslq", "gaussmle"] = "gausslq",
    modality: Literal["astigmatic"] = "astigmatic",
) -> np.ndarray:
    """Calculate axial localization precision for given localizations
    based on calibration.

    Parameters
    ----------
    locs : np.recarray
        Localizations.
    info : list of dicts
        Localizations metadata.
    calibration : dict
        Calibration dictionary with x and y coefficients, z step size
        and the number of frames.
    fitting_method : {"gausslq", "gaussmle"}, optional
        Fitting method used to obtain 2D localization parameters (x, y,
        sx, sy). Default is "gausslq".
    modality : {"astigmatic"}, optional
        3D imaging modality. Currently, only "astigmatic" is supported.
        Default is "astigmatic".

    Returns
    -------
    lpz: np.ndarray
        Calculated lpz values for the given localizations in nm.
    """
    if modality != "astigmatic":
        raise NotImplementedError(
            "Currently only 'astigmatic' modality is supported."
        )
    lpz = axial_localization_precision_astig(
        locs, info, calibration, fitting_method
    )
    return lpz


def axial_localization_precision_astig(
    locs: np.recarray,
    info: list[dict],
    calibration: dict,
    fitting_method: Literal["gausslq", "gaussmle"] = "gausslq",
) -> np.ndarray:
    """Calculate axial localization precision for astigmatic 3D imaging
    for given localizations based on calibration.

    Based on Kowalewski, Reinhardt, et al. *DOI will be added once available*.

    Parameters
    ----------
    locs : np.recarray
        Localizations.
    info : list of dicts
        Localizations metadata.
    calibration : dict
        Calibration dictionary with x and y coefficients, z step size
        and the number of frames.
    fitting_method : {"gausslq", "gaussmle"}, optional
        Fitting method used to obtain 2D localization parameters (x, y,
        sx, sy). Default is "gausslq".

    Returns
    -------
    lpz: np.ndarray
        Calculated lpz values for the given localizations in nm.
    """
    assert fitting_method in [
        "gausslq",
        "gaussmle",
    ], "fitting_method must be 'gausslq' or 'gaussmle'."
    assert (
        "X Coefficients" in calibration
        and "Y Coefficients" in calibration
        and "Magnification factor" in calibration
    ), "Calibration dictionary must contain 'X Coefficients', 'Y Coefficients', and 'Magnification factor'."

    # get camera pixel size
    pixelsize = lib.get_from_metadata(info, "Pixelsize")
    if pixelsize is None:
        raise ValueError("Pixelsize not found in info.")

    mag_factor = calibration["Magnification factor"]
    cx = np.array(calibration["X Coefficients"])
    cy = np.array(calibration["Y Coefficients"])
    lpz = _axial_localization_precision_astig(
        locs, cx, cy, mag_factor, pixelsize, fitting_method
    )
    return lpz


def _axial_localization_precision_astig(
    locs: pd.DataFrame,
    cx: np.ndarray,
    cy: np.ndarray,
    magnification_factor: float,
    pixelsize: float,
    fitting_method: Literal["gausslq", "gaussmle"] = "gausslq",
) -> np.ndarray:
    """Calculate axial localization precision for astigmatic 3D imaging
    for given localizations based on calibration.

    Based on Kowalewski, Reinhardt, et al. *DOI will be added once available*.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations. Must include columns 'photons', 'sx', 'sy', 'bg',
        and 'z', see https://picassosr.readthedocs.io/en/latest/files.html#localization-hdf5-files
    cx : np.ndarray
        3D calibration coefficients for x.
    cy : np.ndarray
        3D calibration coefficients for y.
    pixelsize : float
        Camera pixel size in nm.
    fitting_method : {"gausslq", "gaussmle"}, optional
        Fitting method used to obtain 2D localization parameters (x, y,
        sx, sy). Default is "gausslq".

    Returns
    -------
    lpz: np.ndarray
        Calculated lpz values for the given localizations in nm.
    """
    if fitting_method == "gausslq":
        sigma_uncertainty = gausslq.sigma_uncertainty
    elif fitting_method == "gaussmle":
        sigma_uncertainty = gaussmle.sigma_uncertainty
    else:
        raise ValueError("fitting_method must be 'gausslq' or 'gaussmle'.")

    # to pinpoint what was the actual spot size during measurement
    z = locs["z"] / magnification_factor
    wx_calib = get_calib_size(cx, z) * pixelsize
    wy_calib = get_calib_size(cy, z) * pixelsize
    wx_calib_prime = get_prime_calib_size(cx, z) * pixelsize
    wy_calib_prime = get_prime_calib_size(cy, z) * pixelsize
    sqrt_wx_calib = np.sqrt(wx_calib)
    sqrt_wx_calib_prime = wx_calib_prime / (2 * sqrt_wx_calib)
    sqrt_wy_calib = np.sqrt(wy_calib)
    sqrt_wy_calib_prime = wy_calib_prime / (2 * sqrt_wy_calib)
    se_sx = (
        sigma_uncertainty(locs["sx"], locs["sy"], locs["photons"], locs["bg"])
        * pixelsize
    )
    se_sy = (
        sigma_uncertainty(locs["sy"], locs["sx"], locs["photons"], locs["bg"])
        * pixelsize
    )
    delta_sqrt_wx = (1 / (2 * np.sqrt(locs["sx"] * pixelsize))) * se_sx
    delta_sqrt_wy = (1 / (2 * np.sqrt(locs["sy"] * pixelsize))) * se_sy
    swxc2 = sqrt_wx_calib_prime**2
    swyc2 = sqrt_wy_calib_prime**2
    swx2 = delta_sqrt_wx**2
    swy2 = delta_sqrt_wy**2
    lpz = np.sqrt((swxc2 * swx2 + swyc2 * swy2) / (swxc2 + swyc2) ** 2)
    return lpz * magnification_factor


def get_calib_size(coeffs: np.ndarray, z: np.ndarray) -> np.ndarray:
    """Calculate calibration spot size at the given z position given
    the calibration coefficients. Based on Huang et al., Science 2008."""
    size = (
        coeffs[0] * z**6
        + coeffs[1] * z**5
        + coeffs[2] * z**4
        + coeffs[3] * z**3
        + coeffs[4] * z**2
        + coeffs[5] * z
        + coeffs[6]
    )
    return size


def get_prime_calib_size(coeffs: np.ndarray, z: np.ndarray) -> np.ndarray:
    """Same as ``get_calib_size`` but for the derivative of the size
    function."""
    size_prime = (
        6 * coeffs[0] * z**5
        + 5 * coeffs[1] * z**4
        + 4 * coeffs[2] * z**3
        + 3 * coeffs[3] * z**2
        + 2 * coeffs[4] * z
        + coeffs[5]
    )
    return size_prime



import streamlit as st
from helper import _db_filename
from sqlalchemy import create_engine
import pandas as pd
import os
import numpy as np
from picasso import io
import plotly.express as px


@st.cache_data
def load_file(path: str):
    """Loads a localization file and returns as pandas Dataframe.
    Adds a column with the filename.

    Args:
        path (str): Path to localization file.
        file (str): filename
    """

    locs, info = io.load_locs(path)
    locs = pd.DataFrame(locs)
    locs["file"] = os.path.split(path)[-1]
    return locs, info


def get_file_family(file: str):
    """Returns all files that belong to a parent file for picasso.
    E.g. for a folder with 'file.hdf5' and 'file_undrift.hdf5',
    when searching for 'file.hdf5', both files will be returned.

    Args:
        file (str): Path to file.
    """
    base = os.path.split(file)[1].split(".")[0]
    folder = os.path.dirname(file)

    files = os.listdir(folder)
    files = [f for f in files if f.startswith(base) and f.endswith(".hdf5")]

    return files, folder


def locs_per_frame_plot(hdf_dict: dict):
    """Plots the localizations per frame.

    Args:
        hdf_dict (dict): Dictionary with hdf summary information
    """
    smooth = st.number_input("Smooth", value=100, min_value=1, max_value=1000)
    summary = []

    for f, df in hdf_dict.items():
        d = df[["frame", "photons"]].groupby("frame").count()
        d.columns = ["count"]
        d = d.rolling(smooth).mean()
        d["file"] = f
        d = d.reset_index()

        summary.append(d)

    plot_df = pd.concat(summary, axis=0)

    fig = px.line(
        plot_df,
        x="frame",
        y="count",
        title="Locs per Frame",
        color="file",
        height=600,
    )

    fig.update_layout(
        legend=dict(
            x=0,
            y=-0.5,
        )
    )

    st.plotly_chart(fig)


def hist_plot(hdf_dict: dict, locs: pd.DataFrame):
    """Plots a histogram for a given hdf dictionary.

    Args:
        hdf_dict (dict): Dictionary with summary stats per file.
        locs (pd.DataFrame): pandas Dataframe with localizations.
    """

    c1, c2, c3, c4 = st.columns(4)

    fields = locs.columns
    field = c1.selectbox("Select field", fields)

    try:
        n_bins = c2.number_input(
            "Number of bins", value=100, min_value=1, max_value=200
        )

        all_f = {}
        for f, df in hdf_dict.items():
            all_f[f] = df[field].values

        all_values = np.concatenate(list(all_f.values()))

        min_ = float(np.min(all_values))
        max_ = float(np.max(all_values))

        upper = float(np.percentile(all_values, 99))
        lower = float(np.percentile(all_values, 1))

        min_value = c3.number_input(
            "Min value", value=lower, min_value=min_, max_value=max_
        )
        max_value = c4.number_input(
            "Max value", value=upper, min_value=min_value, max_value=max_
        )

        bins = np.linspace(min_value, max_value, n_bins)

        summary = []
        for f, vals in all_f.items():

            counts, _ = np.histogram(vals, bins=bins)
            sub_df = pd.DataFrame([bins[1:] + bins[0] / 2, counts]).T

            sub_df.columns = [field, "count"]
            sub_df["file"] = f
            summary.append(sub_df)

        plot_df = pd.concat(summary, axis=0)

        fig = px.line(
            plot_df,
            x=field,
            y="count",
            title=f"{field}",
            color="file",
            height=600,
        )

        fig.update_layout(
            legend=dict(
                x=0,
                y=-0.5,
            )
        )

        st.plotly_chart(fig)
    except Exception as e:
        st.warning(f"An error occured plotting field **{field}**.\n {e}")


def compare():
    """Compare streamlit page."""
    st.write("# Compare")

    st.write(
        "Compare multiple files from the database. All hdf files with the same"
        " base path as the movie will be selectable."
    )

    engine = create_engine("sqlite:///" + _db_filename(), echo=False)

    try:
        df = pd.read_sql_table("files", con=engine)

        files = df["filename"].unique()

        selected = st.multiselect(
            "Select files (Hover to see full path)",
            files,
        )

        if len(selected) > 0:

            file_dict = {}
            hdf_dict = {}
            for file in selected:
                try:
                    c1, f1 = get_file_family(file)
                    file_dict[file] = st.multiselect(
                        f"Select hdf file for {file}",
                        c1,
                        None,
                    )

                    if file_dict[file] is not None:
                        for _ in file_dict[file]:
                            path = os.path.dirname(file)

                            locs_filename = os.path.join(path, _)

                            with st.spinner("Loading files"):
                                locs, info = load_file(locs_filename)
                                hdf_dict[locs_filename] = locs
                except FileNotFoundError:
                    st.error(
                        f"File **{file}** was not found. Please check that "
                        "this file exists."
                    )

            st.write("## Plot")

            if len(hdf_dict) > 0:
                with st.spinner("Generating plots"):
                    plot = st.selectbox(
                        "Select plot", ["Localizations per frame", "Histogram"]
                    )
                    if plot == "Localizations per frame":
                        locs_per_frame_plot(hdf_dict)
                    else:
                        hist_plot(hdf_dict, locs)
            else:
                st.warning("Please select HDF files.")

    except ValueError:
        st.warning("Database empty. Process files first.")



"""Test picasso.localize functions as well as the associated functions
in picasso.gausslq, picasso.gaussmle, picasso.zfit

:author: Rafal Kowalewski, 2025
:copyright: Copyright (c) 2025 Jungmann Lab, MPI of Biochemistry
"""

# TODO: add identifying more data types? like .tif, .nd2? this can go to test_io.py though
# TODO: add calibration tests

import numpy as np
import pytest
from picasso import io, localize, gausslq, gaussmle, zfit

# parameters for localization
BOX = 7
MIN_NG = 5000
CAMERE_INFO = {
    "Baseline": 0,
    "Sensitivity": 1,
    "Gain": 1,
}
ROI = ((0, 0), (16, 32))
DRIFT_SEG = 100
CALIB_3D = {
    "X Coefficients": [
        -1.6680708772714857e-18,
        2.4038209829154137e-15,
        2.1771067332017187e-12,
        -3.0324788231238476e-09,
        3.5433326085494675e-06,
        0.0023039289366630425,
        1.2026032603707493,
    ],
    "Y Coefficients": [
        -1.7708672355491796e-18,
        9.808249540501714e-16,
        2.10653248543535e-12,
        2.228026137415219e-11,
        3.628007433361433e-06,
        -0.001646865504353452,
        1.2257249554338714,
    ],
}


@pytest.fixture(scope="module")
def movie_data():
    """Load movie data once per test module."""
    movie, info = io.load_movie("./tests/data/testdata.raw")
    return movie, info


@pytest.fixture
def movie(movie_data):
    """Provide movie data to test functions."""
    return movie_data[0]


@pytest.fixture
def info(movie_data):
    """Provide info data to test functions."""
    return movie_data[1]


@pytest.fixture
def identifications(movie):
    """Provide identifications to test functions."""
    ids = localize.identify(movie, MIN_NG, BOX)
    return ids


@pytest.fixture
def spots(movie, identifications):
    """Provide extracted spots to test functions."""
    spots = localize.get_spots(movie, identifications, BOX, CAMERE_INFO)
    return spots


@pytest.fixture
def theta_lq(spots):
    """Provide least-squares fitting results."""
    return gausslq.fit_spots(spots)


@pytest.fixture
def locs_lq(identifications, theta_lq):
    """Provide 2D localizations from least-squares fitting."""
    return gausslq.locs_from_fits(identifications, theta_lq, BOX, False)


def test_identification_with_roi(movie, identifications):
    """Test identification of spots in movie frames.

    Parameters
    ----------
    movie : np.memmap or np.ndarray
        Movie data as 3D array (frames, height, width)
    info : dict
        Movie metadata information

    Returns
    -------
    pd.DataFrame
        DataFrame containing identified spots with columns:
        'frame', 'x', 'y', 'net_gradient'
    """
    # run identification with the defined parameters
    identifications_roi = localize.identify(movie, MIN_NG, BOX, roi=ROI)

    # basic validation tests
    assert not identifications.empty, "No spots were identified"
    assert all(
        col in identifications.columns
        for col in ["frame", "x", "y", "net_gradient"]
    ), "Missing required columns in identification results"
    assert (
        identifications["net_gradient"].min() >= MIN_NG
    ), f"Found spots with net_gradient below minimum threshold {MIN_NG}"
    assert identifications["frame"].min() >= 0, "Invalid frame numbers found"
    assert identifications["frame"].max() < len(
        movie
    ), "Frame numbers exceed movie length"

    # compare the two identifications
    assert len(identifications_roi) <= len(
        identifications
    ), "ROI identification returned more spots than full frame identification"


def test_identification_threaded_vs_non_threaded(movie, identifications):
    """Test that threaded and non-threaded identification give
    consistent results."""
    # get results from both modes
    ids_threaded = localize.identify(movie, MIN_NG, BOX, threaded=True)

    # compare results (should be identical or very similar)
    assert len(ids_threaded) == len(
        identifications
    ), "Different number of spots identified in threaded vs non-threaded mode"


def test_localize_lq(identifications, spots, theta_lq):
    """Test localizing identified spots using least-squares fitting,
    including their preprocessing."""
    # test extracting spots
    assert len(spots) == len(
        identifications
    ), "Number of extracted spots does not match number of identifications"

    # test localization via least-squares fitting
    theta_lq_multi = gausslq.fit_spots_parallel(spots, asynch=False)
    assert len(theta_lq) == len(
        spots
    ), "Number of localized spots (LQ) does not match number of extracted spots"
    assert len(theta_lq_multi) == len(
        spots
    ), "Number of localized spots (LQ multi) does not match number of extracted spots"
    assert np.allclose(
        theta_lq, theta_lq_multi
    ), "LQ fitting results differ between single-threaded and multi-threaded implementations"


def test_localize_mle(identifications, spots):
    """Test localizing identified spots via maximum-likelihood
    fitting."""
    # test localization via maximum-likelihood fitting
    theta_mle_xy, CRLBs, lls, its = gaussmle.gaussmle(
        spots, eps=1e-3, max_it=1000, method="sigmaxy"
    )
    theta_mle, _, _, _ = gaussmle.gaussmle(
        spots, eps=1e-3, max_it=1000, method="sigma"
    )
    locs_mle = gaussmle.locs_from_fits(
        identifications, theta_mle_xy, CRLBs, lls, its, BOX
    )
    assert len(theta_mle_xy) == len(
        spots
    ), "Number of localized spots (MLE sigmaxy) does not match number of extracted spots"
    assert len(theta_mle) == len(
        spots
    ), "Number of localized spots (MLE sigma) does not match number of extracted spots"

    # TODO: the interface for mle fitting via parallel processing is a
    # little messy, needs to be made more analogous to the lq fitting,
    # then the tests can be added here as well.


def test_localize_3d(info, locs_lq):
    """Test 3D localization of identified spots."""
    locs_3d = zfit.fit_z(locs_lq, info, CALIB_3D, 0.79, 130)
    assert "z" in locs_3d.columns, "3D localization results missing 'z' column"
    assert len(locs_3d), "No 3D localized spots were returned"
    locs_3d_multi = zfit.fit_z_parallel(
        locs_lq, info, CALIB_3D, 0.79, 130, asynch=False
    )
    assert len(locs_3d) == len(
        locs_lq
    ), "Number of 3D localized spots does not match number of 2D localized spots"
    assert len(locs_3d_multi) == len(
        locs_lq
    ), "Number of 3D localized spots (multi) does not match number of 2D localized spots"



"""
picasso.imageprocess
~~~~~~~~~~~~~~~~~~~~

Image processing functions

:author: Joerg Schnitzbauer, 2016
:copyright: Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry
"""

from __future__ import annotations

from typing import Callable

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from tqdm import tqdm

from . import lib, render, localize, postprocess


plt.style.use("ggplot")


def xcorr(imageA: np.ndarray, imageB: np.ndarray) -> np.ndarray:
    """Compute the cross-correlation of two images using FFT.

    Parameters
    ----------
    imageA, imageB : np.ndarray
        Input images to be cross-correlated. They should have the same
        shape.

    Returns
    -------
    res : np.ndarray
        The cross-correlation result, which is the inverse Fourier
        transform of the product of the Fourier transforms of the two
        images.
    """
    FimageA = np.fft.fft2(imageA)
    CFimageB = np.conj(np.fft.fft2(imageB))
    res = np.fft.fftshift(
        np.real(np.fft.ifft2((FimageA * CFimageB)))
    ) / np.sqrt(imageA.size)
    return res


def get_image_shift(
    imageA: np.ndarray,
    imageB: np.ndarray,
    box: int,
    roi: int | None = None,
    display: bool = False,
) -> tuple[float, float]:
    """Compute the shift from ``imageA`` to ``imageB``.

    Parameters
    ----------
    imageA, imageB : np.ndarray
        Input images to be cross-correlated. They should have the same
        shape.
    box : int
        Size of the box used for fitting the cross-correlation peak.
    roi : int, optional
        Region of interest size to cut out the center of the
        cross-correlation image. If None, the entire cross-correlation
        image is used.
    display : bool, optional
        If True, displays the images and the cross-correlation result.

    Returns
    -------
    yc, xc : float
        The y and x coordinates of the shift from imageA to imageB.
        The coordinates are adjusted to be relative to the center of
        the images.
    """
    if (np.sum(imageA) == 0) or (np.sum(imageB) == 0):
        return 0, 0
    # Compute image correlation
    XCorr = xcorr(imageA, imageB)
    # Cut out center roi
    Y, X = imageA.shape
    if roi is not None:
        Y_ = int((Y - roi) / 2)
        X_ = int((X - roi) / 2)
        if Y_ > 0:
            XCorr = XCorr[Y_:-Y_, :]
        else:
            Y_ = 0
        if X_ > 0:
            XCorr = XCorr[:, X_:-X_]
        else:
            X_ = 0
    else:
        Y_ = X_ = 0
    # A quarter of the fit ROI
    fit_X = int(box / 2)
    # A coordinate grid for the fitting ROI
    y, x = np.mgrid[-fit_X : fit_X + 1, -fit_X : fit_X + 1]
    # Find the brightest pixel and cut out the fit ROI
    y_max_, x_max_ = np.unravel_index(XCorr.argmax(), XCorr.shape)
    FitROI = XCorr[
        y_max_ - fit_X : y_max_ + fit_X + 1,
        x_max_ - fit_X : x_max_ + fit_X + 1,
    ]

    dimensions = FitROI.shape

    if 0 in dimensions or dimensions[0] != dimensions[1]:
        xc, yc = 0, 0
    else:

        def flat_2d_gaussian(coords, a, xc, yc, s, b):
            x, y = coords
            A = a * np.exp(-0.5 * ((x - xc) ** 2 + (y - yc) ** 2) / s**2) + b
            return A.flatten()

        p0 = [FitROI.max(), 0, 0, 1, FitROI.min()]
        bounds = (
            [0, -np.inf, -np.inf, 0, 0],
            [np.inf, np.inf, np.inf, np.inf, np.inf],
        )
        popt, _ = curve_fit(
            flat_2d_gaussian,
            (x, y),
            FitROI.flatten(),
            p0=p0,
            bounds=bounds,
        )

        # Get maximum coordinates and add offsets
        xc = popt[1]
        yc = popt[2]
        xc += X_ + x_max_
        yc += Y_ + y_max_

        if display:
            plt.figure(figsize=(17, 10), constrained_layout=True)
            plt.subplot(1, 3, 1)
            plt.imshow(imageA, interpolation="none")
            plt.subplot(1, 3, 2)
            plt.imshow(imageB, interpolation="none")
            plt.subplot(1, 3, 3)
            plt.imshow(XCorr, interpolation="none")
            plt.plot(xc, yc, "x")
            plt.show()

        xc -= np.floor(X / 2)
        yc -= np.floor(Y / 2)

    return -yc, -xc


def rcc(
    segments: list[np.ndarray],
    max_shift: float | None = None,
    callback: Callable[[int], None] | None = None,
) -> tuple[np.ndarray, np.ndarray]:
    """Compute RCC, see Wang, Schnitzbauer, et al. Optics Express,
    2014. Return the shifts in x and y directions for each pair of
    segments.

    Parameters
    ----------
    segments : list of np.ndarray
        List of image segments to be correlated. Each segment should be
        a 2D numpy array representing an image.
    max_shift : float, optional
        Maximum allowed shift in pixels. If None, the default value is
        set to 5 pixels.
    callback : Callable[[int], None], optional
        A callback function that takes an integer argument and is called
        after processing each pair of segments. This can be used to
        update a progress bar or perform other actions during the
        computation.

    Returns
    -------
    shifts_x, shifts_y : np.ndarray
        2D numpy arrays containing the shifts in x and y directions for
        each pair of segments. The shape of the arrays is
        (n_segments, n_segments), where n_segments is the number of
        segments provided.
    """
    n_segments = len(segments)
    shifts_x = np.zeros((n_segments, n_segments))
    shifts_y = np.zeros((n_segments, n_segments))
    n_pairs = int(n_segments * (n_segments - 1) / 2)
    flag = 0
    if callback is None:
        with tqdm(
            total=n_pairs, desc="Correlating image pairs", unit="pairs"
        ) as progress_bar:
            for i in range(n_segments - 1):
                for j in range(i + 1, n_segments):
                    progress_bar.update()
                    shifts_y[i, j], shifts_x[i, j] = get_image_shift(
                        segments[i], segments[j], 5, max_shift
                    )
                    flag += 1
    else:
        callback(0)
        for i in range(n_segments - 1):
            for j in range(i + 1, n_segments):
                shifts_y[i, j], shifts_x[i, j] = get_image_shift(
                    segments[i], segments[j], 5, max_shift
                )
                flag += 1
                callback(flag)

    return lib.minimize_shifts(shifts_x, shifts_y)


def find_fiducials(
    locs: pd.DataFrame,
    info: list[dict],
) -> tuple[list[tuple[int, int]], int]:
    """Find the xy coordinates of regions with high density of
    localizations, likely originating from fiducial markers.

    Uses ``picasso.localize.identify_in_image`` with threshold set to
    99th percentile of the image histogram. The image is rendered using
    one-pixel-blur, see picasso.render.render.

    Parameters
    ----------
    locs : pd.DataFrame
        Localizations.
    info : list of dicts
        Localizations' metadata (from the corresponding .yaml file).

    Returns
    -------
    picks : list of (2,) tuples
        Coordinates of fiducial markers. Each list element corresponds
        to (x, y) coordinates of one fiducial marker.
    box : int
        Size of the box used for the fiducial marker identification.
        Can be set as the pick diameter in pixels for undrifting.
    """
    image = render.render(
        locs=locs,
        info=info,
        oversampling=1,
        viewport=None,
        blur_method="smooth",
    )[1]
    # hist = np.histogram(image.flatten(), bins=256)
    threshold = np.percentile(image.flatten(), 99)
    # box size should be an odd number, corresponding to approximately
    # 900 nm
    pixelsize = lib.get_from_metadata(info, "Pixelsize", default=130)
    box = int(np.round(900 / pixelsize))
    box = box + 1 if box % 2 == 0 else box

    # find the local maxima and translate to pick coordinates
    y, x, _ = localize.identify_in_image(image, threshold, box=box)
    picks = [(xi, yi) for xi, yi in zip(x, y)]

    # select the picks with appropriate number of localizations
    n_frames = lib.get_from_metadata(info, "Frames", default=0)
    min_n = 0.8 * n_frames
    picked_locs = postprocess.picked_locs(
        locs,
        info,
        picks,
        "Circle",
        pick_size=box / 2,
        add_group=False,
    )
    picks = [
        pick for i, pick in enumerate(picks) if len(picked_locs[i]) > min_n
    ]
    return picks, box


def radial_sum(image: np.ndarray) -> np.ndarray:
    """Compute the radial projection of the sum of pixel values.

    If the radial distance of a pixel to center is r, then the sum of the
    intensities of all pixels with n <= r < (n + 1) is stored at
    position n in the output array.

    Parameters
    ----------
    image : np.ndarray
        Input image array.

    Returns
    -------
    radial_profile : np.ndarray
        1D array containing the radial sum values.

    Raises
    ------
    ValueError
        If ``max_radius`` is not "inner_radius" or "outer_radius".
    """
    assert image.ndim == 2, "Input image must be 2D."
    assert image.shape[0] == image.shape[1], "Input image must be square."
    assert image.shape[0] % 2 == 1, "Input image size must be odd."

    size = image.shape[0]
    center = size // 2
    r = np.arange(0, center + 1)
    # get the square distances of each pixel from the center
    y, x = np.ogrid[:size, :size]
    dist_sq = (x - center) ** 2 + (y - center) ** 2
    # create an array to hold the counts
    counts = np.zeros_like(r, dtype=image.dtype)
    # iterate over each radius and compute the sum and count
    for r_idx, radius in enumerate(r):
        mask = (dist_sq >= radius**2) & (dist_sq < (radius + 1) ** 2)
        counts[r_idx] = np.sum(image[mask])
    return counts



"""
ext/bitplane
~~~~~~~~~~~~
Utility functions to handle bitplane data.
:author: Maximilian T Strauss, 2021-2022
:copyright: Copyright (c) 2021-2022 Maximilian T Strauss
"""

import os.path as _ospath
import numpy as np
import pandas as pd
import datetime

# 0.9.6 installer has had problems with PyImarisWriter, TODO: fix later
try:
    # from PyImarisWriter.ImarisWriterCtypes import *
    from PyImarisWriter import PyImarisWriter as PW

    IMSWRITER = True
except ModuleNotFoundError:
    IMSWRITER = False

if IMSWRITER:

    class MovieMapper:
        """
        MovieMapper class to map ims files.
        """

        def __init__(
            self,
            file,
            RL,
            lookup_dict,
            channel,
            frames,
            x,
            y,
            dtype,
        ):
            self.file = file
            self.RL = RL
            self.lookup_dict = lookup_dict
            self.channel = channel
            self.x = x
            self.y = y
            self.frames = frames
            self.dtype = dtype

        def __getitem__(self, item):
            return self.file["DataSet"][self.RL][self.lookup_dict[item]][
                self.channel
            ]["Data"][0][: self.y, : self.x]

        def __len__(self):
            return len(self.frames)

        def __iter__(self):
            for item in range(len(self.frames)):
                yield self.file["DataSet"][self.RL][self.lookup_dict[item]][
                    self.channel
                ]["Data"][0][: self.y, : self.x]

    class MovieMapperStack:
        """
        MovieMapperStack class to map ims files. This class is for z-stacks.
        """

        def __init__(self, file, RL, channel, frames, dtype, n_frames):
            self.file = file
            self.RL = RL
            self.channel = channel
            self.frames = frames
            self.dtype = dtype
            self.n_frames = n_frames

        def __getitem__(self, item):
            return self.file["DataSet"][self.RL][self.frames[0]][self.channel][
                "Data"
            ][item, :, :]

        def __len__(self):
            return self.n_frames

        def __iter__(self):
            for item in range(self.n_frames):
                yield self.file["DataSet"][self.RL][self.frames[0]][
                    self.channel
                ]["Data"][item, :, :]

    class IMSFile:
        """
        Class for handling IMS files.

        This class is assuming the following:
            * ResolutionLevel 0 is the default.
            * The image size doesn't change from one channel to another.

        TODO:
            * check if channel exists..

        """

        RL = "ResolutionLevel 0"

        def __init__(self, path, verbose=False):
            if verbose:
                print("Reading info from {}".format(path))
            self.path = _ospath.abspath(path)
            self.file = pd.read_hdf(self.path)

            self.frames = list(self.file["DataSet"][self.RL].keys())
            self.n_frames = len(self.frames)
            self.channels = list(
                self.file["DataSet"][self.RL][self.frames[0]].keys()
            )

            self.set_channel(self.channels[0])

            self.frames_read = False
            self.lookup_dict = None

        def set_channel(self, channel):
            self.channel = channel
            self.img_size = np.array(
                self.file["DataSet"][self.RL][self.frames[0]][self.channel][
                    "Data"
                ]
            ).shape
            self.dtype = np.array(
                self.file["DataSet"][self.RL][self.frames[0]][self.channel][
                    "Data"
                ]
            ).dtype

            try:
                z = int(
                    "".join(
                        [
                            _.decode()
                            for _ in self.file["DataSetInfo"]["Image"].attrs[
                                "Z"
                            ]
                        ]
                    )
                )
            except KeyError:
                z = self.img_size[0]
            self.z = z

            try:
                self.x = int(
                    "".join(
                        [
                            _.decode()
                            for _ in self.file["DataSetInfo"]["Image"].attrs[
                                "X"
                            ]
                        ]
                    )
                )
                self.y = int(
                    "".join(
                        [
                            _.decode()
                            for _ in self.file["DataSetInfo"]["Image"].attrs[
                                "Y"
                            ]
                        ]
                    )
                )
            except KeyError:
                self.x = self.img_size[1]
                self.y = self.img_size[2]

            # The pixelsize is being estimated on the image dimensions

            ext_max0 = float(
                "".join(
                    [
                        _.decode()
                        for _ in self.file["DataSetInfo"]["Image"].attrs[
                            "ExtMax0"
                        ]
                    ]
                )
            )
            ext_min0 = float(
                "".join(
                    [
                        _.decode()
                        for _ in self.file["DataSetInfo"]["Image"].attrs[
                            "ExtMin0"
                        ]
                    ]
                )
            )

            ext_max1 = float(
                "".join(
                    [
                        _.decode()
                        for _ in self.file["DataSetInfo"]["Image"].attrs[
                            "ExtMax1"
                        ]
                    ]
                )
            )
            ext_min1 = float(
                "".join(
                    [
                        _.decode()
                        for _ in self.file["DataSetInfo"]["Image"].attrs[
                            "ExtMin1"
                        ]
                    ]
                )
            )

            ext_max2 = float(
                "".join(
                    [
                        _.decode()
                        for _ in self.file["DataSetInfo"]["Image"].attrs[
                            "ExtMax2"
                        ]
                    ]
                )
            )
            ext_min2 = float(
                "".join(
                    [
                        _.decode()
                        for _ in self.file["DataSetInfo"]["Image"].attrs[
                            "ExtMin2"
                        ]
                    ]
                )
            )

            delta_x = ext_max0 - ext_min0
            delta_y = ext_max1 - ext_min1

            px_x = delta_x / self.x * 1000
            px_y = delta_y / self.y * 1000

            px_nm = (px_x + px_y) / 2

            print(
                f"Image dimensions X: {delta_x} um Y: {delta_y} um. "
                f"Pixelsize x {px_x}, y {px_y}, p {px_nm}."
            )

            self.pixelsize = px_nm

            self.ext_max0 = ext_max0
            self.ext_max1 = ext_max1
            self.ext_max2 = ext_max2

            self.ext_min0 = ext_min0
            self.ext_min1 = ext_min1
            self.ext_min2 = ext_min2

        def read_frames(self):
            if not self.frames_read:
                self.frames_int = np.array(
                    [int(_.split("TimePoint ")[1]) for _ in self.frames]
                )
                self.lookup_dict = dict(zip(self.frames_int, self.frames))

                self.frames_read = True

        def read_frame(self, frame):
            if not self.lookup_dict:
                self.read_frames()
            return self.file["DataSet"][self.RL][self.lookup_dict[frame]][
                self.channel
            ]["Data"][0]

        def read_z_stack(self):
            print("Reading stack")
            self.n_frames = self.z
            self.movie = MovieMapperStack(
                self.file,
                self.RL,
                self.channel,
                self.frames,
                self.dtype,
                self.n_frames,
            )

        def read_stack(self):
            print("Reading movie")
            if not self.lookup_dict:
                self.read_frames()

            self.movie = MovieMapper(
                self.file,
                self.RL,
                self.lookup_dict,
                self.channel,
                self.frames,
                self.x,
                self.y,
                self.dtype,
            )

        def read_movie(self):

            if self.z > 1:
                self.read_z_stack()
            else:
                self.read_stack()

    class ImsCallback(PW.CallbackClass):
        def __init__(self):
            self.mUserDataProgress = 0

        def RecordProgress(self, progress, total_bytes_written):
            progress100 = int(progress * 100)
            if progress100 - self.mUserDataProgress >= 5:
                self.mUserDataProgress = progress100
                # print(
                #     "User Progress {}, Bytes written: {}".format(
                #         self.mUserDataProgress, total_bytes_written
                #     )
                # )

    def numpy_to_imaris(
        array,
        filename,
        colors,
        oversampling,
        viewport,
        info,
        z_min,
        z_max,
        pixelsize,
    ):

        if len(array.shape) == 3:
            array = np.expand_dims(array, 1)

        dtype = array.dtype
        dtype_imaris = str(dtype)

        x = array.shape[3]
        y = array.shape[2]
        z = array.shape[1]
        c = array.shape[0]

        x_min = viewport[0][1]
        x_max = viewport[1][1]
        y_min = viewport[0][0]
        y_max = viewport[1][0]

        np_data = array.copy()

        image_size = PW.ImageSize(x=x, y=y, z=z, c=c, t=1)
        dimension_sequence = PW.DimensionSequence("x", "y", "z", "c", "t")
        block_size = image_size
        sample_size = PW.ImageSize(x=1, y=1, z=1, c=1, t=1)
        output_filename = filename

        options = PW.Options()
        options.mNumberOfThreads = 1
        options.mCompressionAlgorithmType = PW.eCompressionAlgorithmGzipLevel2
        options.mEnableLogProgress = True

        application_name = "PyImarisWriter"
        application_version = "1.0.0"

        callback_class = ImsCallback()
        converter = PW.ImageConverter(
            dtype_imaris,
            image_size,
            sample_size,
            dimension_sequence,
            block_size,
            output_filename,
            options,
            application_name,
            application_version,
            callback_class,
        )

        num_blocks = image_size / block_size

        block_index = PW.ImageSize()
        for c in range(num_blocks.c):
            block_index.c = c
            for t in range(num_blocks.t):
                block_index.t = t
                for z in range(num_blocks.z):
                    block_index.z = z
                    for y in range(num_blocks.y):
                        block_index.y = y
                        for x in range(num_blocks.x):
                            block_index.x = x
                            if converter.NeedCopyBlock(block_index):
                                converter.CopyBlock(np_data, block_index)

        adjust_color_range = True

        x_0 = (x_min) * pixelsize / 1000
        y_0 = (y_min) * pixelsize / 1000

        x_1 = (x_max) * pixelsize / 1000
        y_1 = (y_max) * pixelsize / 1000

        # TODO: Later use GlobalExtMin to add
        # Todo: Check for z
        try:
            x_0 += info[0]["ExtMin0"]
            y_0 += info[0]["ExtMin1"]

            x_1 += info[0]["ExtMin0"]
            y_1 += info[0]["ExtMin1"]

        except KeyError as e:
            print(f"Exception: {e}")

        try:
            z_base = (info[0]["ExtMin2"] + info[0]["ExtMax2"]) / 2
        except KeyError as e:
            print(f"Exception: {e}")
            z_base = 0

        if z_min == z_max == 0:
            z_0 = z_base - (image_size.z / 2) * pixelsize / 1000 / oversampling
            z_1 = z_base + (image_size.z / 2) * pixelsize / 1000 / oversampling
        else:
            z_0 = z_base + z_min / 1000
            z_1 = z_base + z_max / 1000

        # The image extends are not for multiple z-stacks.
        image_extents = PW.ImageExtents(x_0, y_0, z_0, x_1, y_1, z_1)

        print(f"Image dimensions {x_0, y_0, z_0, x_1, y_1, z_1}\n")

        parameters = PW.Parameters()
        parameters.set_value("Image", "Info", "PicassoExport")
        time_infos = [datetime.datetime.today()]

        # Channels
        color_infos = [PW.ColorInfo() for _ in range(image_size.c)]

        for idx, color in enumerate(colors):
            parameters.set_channel_name(idx, color)
            color_infos[idx].set_base_color(color)

        converter.Finish(
            image_extents,
            parameters,
            time_infos,
            color_infos,
            adjust_color_range,
        )
        converter.Destroy()
        print("Wrote {} to {}".format("Minimal Example", output_filename))



"""
picasso.gaussmle
~~~~~~~~~~~~~~~~

Maximum likelihood fits for single particle localization. Based on
Smith, et al. Nature Methods, 2010.

:authors: Joerg Schnitzbauer, Maximilian Thomas Strauss, 2016-2018
:copyright: Copyright (c) 2016-2018 Jungmann Lab, MPI of Biochemistry
"""

from __future__ import annotations

import math
import multiprocessing
import threading
from concurrent import futures
from typing import Literal

import numba
import numpy as np
import pandas as pd

GAMMA = np.array([1.0, 1.0, 0.5, 1.0, 1.0, 1.0])


@numba.jit(nopython=True, nogil=True)
def _sum_and_center_of_mass(
    spot: np.ndarray,
    size: int,
) -> tuple[float, float, float]:
    """Calculate the sum and center of mass of a 2D spot."""
    y = 0.0
    x = 0.0
    _sum_ = 0.0
    for i in range(size):
        for j in range(size):
            y += spot[i, j] * i
            x += spot[i, j] * j
            _sum_ += spot[i, j]
    y /= _sum_
    x /= _sum_
    return _sum_, y, x


@numba.jit(nopython=True, nogil=True)
def mean_filter(spot: np.ndarray, size: int) -> np.ndarray:
    """Apply a mean filter to the spot. This function computes the mean
    of each pixel in a 3x3 neighborhood.

    Parameters
    ----------
    spot : np.ndarray
        The input image.
    size : int
        The size of the patch (assumed to be square).

    Returns
    -------
    filtered_spot : np.ndarray
        The filtered image patch.
    """
    filtered_spot = np.zeros_like(spot)
    for k in range(size):
        for ll in range(size):
            min_m = np.maximum(0, k - 1)
            max_m = np.minimum(size, k + 2)
            min_n = np.maximum(0, ll - 1)
            max_n = np.minimum(size, ll + 2)
            N = (max_m - min_m) * (max_n - min_n)
            Nsum = 0.0
            for m in range(min_m, max_m):
                for n in range(min_n, max_n):
                    Nsum += spot[m, n]
            filtered_spot[k, ll] = Nsum / N
    return filtered_spot


@numba.jit(nopython=True, nogil=True)
def _initial_sigmas(
    spot: np.ndarray,
    y: None,
    x: None,
    size: int,
) -> tuple[float, float]:
    """Initialize the sizes of the single-emitter images (sigmas of the
    Gaussian fit) in x and y independently."""
    size_half = int(size / 2)
    sum_deviation_y = 0.0
    sum_deviation_x = 0.0
    sum_y = 0.0
    sum_x = 0.0
    for i in range(size):
        d2 = (i - size_half) ** 2
        sum_deviation_y += spot[i, size_half] * d2
        sum_deviation_x += spot[size_half, i] * d2
        sum_y += spot[i, size_half]
        sum_x += spot[size_half, i]
    sy = np.sqrt(sum_deviation_y / sum_y)
    sx = np.sqrt(sum_deviation_x / sum_x)
    if ~np.isfinite(sy):
        sy = 0.01
    if ~np.isfinite(sx):
        sx = 0.01
    if sx == 0:
        sx = 0.01
    if sy == 0:
        sy = 0.01
    return sy, sx


@numba.jit(nopython=True, nogil=True)
def _initial_parameters(
    spot: np.ndarray,
    size: int,
) -> tuple[float, float, float, float, float, float]:
    """Initialize the parameters for the Gaussian fit - x, y, photons,
    background, sigma_x and sigma_y."""
    sum, y, x = _sum_and_center_of_mass(spot, size)
    bg = np.min(mean_filter(spot, size))
    photons = sum - size * size * bg
    photons_sane = np.maximum(1.0, photons)
    sy, sx = _initial_sigmas(spot - bg, y, x, size)
    return x, y, photons_sane, bg, sx, sy


@numba.jit(nopython=True, nogil=True)
def _initial_theta_sigma(spot: np.ndarray, size: int) -> np.ndarray:
    """Initialize the parameters for the Gaussian fit with a single
    sigma for both x and y dimensions - x, y, photons, background,
    sigma."""
    theta = np.zeros(5, dtype=np.float32)
    theta[0], theta[1], theta[2], theta[3], sx, sy = _initial_parameters(
        spot, size
    )
    theta[4] = (sx + sy) / 2
    return theta


@numba.jit(nopython=True, nogil=True)
def _initial_theta_sigmaxy(spot: np.ndarray, size: int) -> np.ndarray:
    """Initialize the parameters for the Gaussian fit with separate
    sigmas for x and y dimensions - x, y, photons, background, sigma_x
    and sigma_y."""
    theta = np.zeros(6, dtype=np.float32)
    theta[0], theta[1], theta[2], theta[3], theta[4], theta[5] = (
        _initial_parameters(spot, size)
    )
    return theta


@numba.vectorize(nopython=True)
def _erf(x: float) -> float:
    """Currently not needed, but might be useful for a CUDA
    implementation."""
    ax = np.abs(x)
    if ax < 0.5:
        t = x * x
        top = (
            (
                (
                    (0.771058495001320e-04 * t - 0.133733772997339e-02) * t
                    + 0.323076579225834e-01
                )
                * t
                + 0.479137145607681e-01
            )
            * t
            + 0.128379167095513e00
        ) + 1.0
        bot = (
            (0.301048631703895e-02 * t + 0.538971687740286e-01) * t
            + 0.375795757275549e00
        ) * t + 1.0
        return x * (top / bot)
    if ax < 4.0:
        top = (
            (
                (
                    (
                        (
                            (-1.36864857382717e-07 * ax + 5.64195517478974e-01)
                            * ax
                            + 7.21175825088309e00
                        )
                        * ax
                        + 4.31622272220567e01
                    )
                    * ax
                    + 1.52989285046940e02
                )
                * ax
                + 3.39320816734344e02
            )
            * ax
            + 4.51918953711873e02
        ) * ax + 3.00459261020162e02
        bot = (
            (
                (
                    (
                        (
                            (1.0 * ax + 1.27827273196294e01) * ax
                            + 7.70001529352295e01
                        )
                        * ax
                        + 2.77585444743988e02
                    )
                    * ax
                    + 6.38980264465631e02
                )
                * ax
                + 9.31354094850610e02
            )
            * ax
            + 7.90950925327898e02
        ) * ax + 3.00459260956983e02
        erf = 0.5 + (0.5 - np.exp(-x * x) * top / bot)
        if x < 0.0:
            erf = -erf
        return erf
    if ax < 5.8:
        x2 = x * x
        t = 1.0 / x2
        top = (
            (
                (2.10144126479064e00 * t + 2.62370141675169e01) * t
                + 2.13688200555087e01
            )
            * t
            + 4.65807828718470e00
        ) * t + 2.82094791773523e-01
        bot = (
            (
                (9.41537750555460e01 * t + 1.87114811799590e02) * t
                + 9.90191814623914e01
            )
            * t
            + 1.80124575948747e01
        ) * t + 1.0
        erf = (0.564189583547756e0 - top / (x2 * bot)) / ax
        erf = 0.5 + (0.5 - np.exp(-x2) * erf)
        if x < 0.0:
            erf = -erf
        return erf
    return np.sign(x)


# TODO: check this is correct?
@numba.jit(nopython=True, nogil=True, cache=False)
def _gaussian_integral(x: float, mu: float, sigma: float) -> float:
    """Calculate the integral of a Gaussian function from negative
    infinity to x."""
    sq_norm = 0.70710678118654757 / sigma  # sq_norm = sqrt(0.5/sigma**2)
    d = x - mu
    return 0.5 * (
        math.erf((d + 0.5) * sq_norm) - math.erf((d - 0.5) * sq_norm)
    )


@numba.jit(nopython=True, nogil=True, cache=False)
def _derivative_gaussian_integral(
    x: float,
    mu: float,
    sigma: float,
    photons: float,
    PSFc: float,
) -> tuple[float, float]:
    """Calculate the first and second derivatives of the integral of a
    Gaussian function with respect to x."""
    d = x - mu
    a = np.exp(-0.5 * ((d + 0.5) / sigma) ** 2)
    b = np.exp(-0.5 * ((d - 0.5) / sigma) ** 2)
    dudt = -photons * PSFc * (a - b) / (np.sqrt(2.0 * np.pi) * sigma)
    d2udt2 = (
        -photons
        * ((d + 0.5) * a - (d - 0.5) * b)
        * PSFc
        / (np.sqrt(2.0 * np.pi) * sigma**3)
    )
    return dudt, d2udt2


@numba.jit(nopython=True, nogil=True, cache=False)
def _derivative_gaussian_integral_1d_sigma(
    x: float,
    mu: float,
    sigma: float,
    photons: float,
    PSFc: float,
) -> tuple[float, float]:
    """Calculate the first and second derivatives of the integral of a
    Gaussian function with respect to sigma in 1D."""
    ax = np.exp(-0.5 * ((x + 0.5 - mu) / sigma) ** 2)
    bx = np.exp(-0.5 * ((x - 0.5 - mu) / sigma) ** 2)
    dudt = (
        -photons
        * (ax * (x + 0.5 - mu) - bx * (x - 0.5 - mu))
        * PSFc
        / (np.sqrt(2.0 * np.pi) * sigma**2)
    )
    d2udt2 = -2.0 * dudt / sigma - photons * (
        ax * (x + 0.5 - mu) ** 3 - bx * (x - 0.5 - mu) ** 3
    ) * PSFc / (np.sqrt(2.0 * np.pi) * sigma**5)
    return dudt, d2udt2


@numba.jit(nopython=True, nogil=True)
def _derivative_gaussian_integral_2d_sigma(
    x: float,
    y: float,
    mu: float,
    nu: float,
    sigma: float,
    photons: float,
    PSFx: float,
    PSFy: float,
) -> tuple[float, float]:
    """Calculate the first and second derivatives of the integral of a
    Gaussian function with respect to sigma in 2D."""
    dSx, ddSx = _derivative_gaussian_integral_1d_sigma(
        x,
        mu,
        sigma,
        photons,
        PSFy,
    )
    dSy, ddSy = _derivative_gaussian_integral_1d_sigma(
        y,
        nu,
        sigma,
        photons,
        PSFx,
    )
    dudt = dSx + dSy
    d2udt2 = ddSx + ddSy
    return dudt, d2udt2


def _worker(
    func,
    spots,
    thetas,
    CRLBs,
    likelihoods,
    iterations,
    eps,
    max_it,
    current,
    lock,
):
    """Worker function for asynchronous Gaussian fitting."""
    N = len(spots)
    while True:
        with lock:
            index = current[0]
            if index == N:
                return
            current[0] += 1
        func(spots, index, thetas, CRLBs, likelihoods, iterations, eps, max_it)


def gaussmle(
    spots: np.ndarray,
    eps: float,
    max_it: int,
    method: Literal["sigma", "sigmaxy"] = "sigmaxy",
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Fits Gaussians using Maximum Likelihood Estimation (MLE) to the
    extracted spots.

    Parameters
    ----------
    spots : np.ndarray
        The input image patches containing the spots of shape
        (N, size, size), where N is the number of spots and size is the
        size of the square patch.
    eps : float
        The convergence criterion for the fitting algorithm.
    max_it : int
        The maximum number of iterations for the fitting algorithm.
    method : Literal["sigma", "sigmaxy"]
        The method to use for fitting the Gaussian.

    Returns
    -------
    thetas : np.ndarray
        The fitted parameters for each spot, shape (N, 6) or (N, 5)
        depending on the method. The columns are x, y, photons,
        background and sigma (or sigmax, sigmay).
    CRLBs : np.ndarray
        The Cramer-Rao Lower Bounds for the fitted parameters, shape
        (N, 6) or (N, 5).
    likelihoods : np.ndarray
        The log-likelihoods for each fitted spot, shape (N,).
    iterations : np.ndarray
        The number of iterations taken to converge for each spot,
        shape (N,).
    """
    N = len(spots)
    thetas = np.zeros((N, 6), dtype=np.float32)
    CRLBs = np.inf * np.ones((N, 6), dtype=np.float32)
    likelihoods = np.zeros(N, dtype=np.float32)
    iterations = np.zeros(N, dtype=np.int32)
    if method == "sigma":
        func = _mlefit_sigma
    elif method == "sigmaxy":
        func = _mlefit_sigmaxy
    else:
        raise ValueError("Method not available.")
    for i in range(N):
        func(spots, i, thetas, CRLBs, likelihoods, iterations, eps, max_it)
    return thetas, CRLBs, likelihoods, iterations


def gaussmle_async(
    spots: np.ndarray,
    eps: float,
    max_it: int,
    method: Literal["sigma", "sigmaxy"] = "sigmaxy",
) -> tuple[list, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Runs ``gaussmle`` asynchronously (multiprocessing) to fit
    Gaussians using Maximum Likelihood Estimation (MLE) to the
    extracted spots. See ``gaussmle`` for parameter details.

    Returns
    -------
    current : np.ndarray
        A single-element array containing the current index of the
        spot being processed.
    thetas, CRLBs, likelihoods, iterations : np.ndarrays
        The same as in ``gaussmle``.
    """
    N = len(spots)
    thetas = np.zeros((N, 6), dtype=np.float32)
    CRLBs = np.inf * np.ones((N, 6), dtype=np.float32)
    likelihoods = np.zeros(N, dtype=np.float32)
    iterations = np.zeros(N, dtype=np.int32)
    n_workers = min(
        60, max(1, int(0.75 * multiprocessing.cpu_count()))
    )  # Python crashes when using >64 cores
    lock = threading.Lock()
    current = [0]
    if method == "sigma":
        func = _mlefit_sigma
    elif method == "sigmaxy":
        func = _mlefit_sigmaxy
    else:
        raise ValueError("Method not available.")
    executor = futures.ThreadPoolExecutor(n_workers)
    for i in range(n_workers):
        executor.submit(
            _worker,
            func,
            spots,
            thetas,
            CRLBs,
            likelihoods,
            iterations,
            eps,
            max_it,
            current,
            lock,
        )
    executor.shutdown(wait=False)
    return current, thetas, CRLBs, likelihoods, iterations


@numba.jit(nopython=True, nogil=True)
def _mlefit_sigma(
    spots: np.ndarray,
    index: int,
    thetas: np.ndarray,
    CRLBs: np.ndarray,
    likelihoods: np.ndarray,
    iterations: np.ndarray,
    eps: float,
    max_it: int,
) -> None:
    """Fits a Gaussian to a single spot using Maximum Likelihood
    Estimation (MLE) with a single sigma for both x and y dimensions."""
    n_params = 5

    spot = spots[index]
    size, _ = spot.shape

    # theta is [x, y, N, bg, S]
    theta = _initial_theta_sigma(spot, size)
    max_step = np.zeros(n_params, dtype=np.float32)
    max_step[0:2] = theta[4]
    max_step[2:4] = 0.1 * theta[2:4]
    max_step[4] = 0.2 * theta[4]

    # Memory allocation
    # (we do that outside of the loops to avoid huge delays in threaded code):
    dudt = np.zeros(n_params, dtype=np.float32)
    d2udt2 = np.zeros(n_params, dtype=np.float32)
    numerator = np.zeros(n_params, dtype=np.float32)
    denominator = np.zeros(n_params, dtype=np.float32)

    old_x = theta[0]
    old_y = theta[1]

    kk = 0
    while (
        kk < max_it
    ):  # we do this instead of a for loop for the special case of max_it=0
        kk += 1

        numerator[:] = 0.0
        denominator[:] = 0.0

        for ii in range(size):
            for jj in range(size):
                PSFx = _gaussian_integral(ii, theta[0], theta[4])
                PSFy = _gaussian_integral(jj, theta[1], theta[4])

                # Derivatives
                dudt[0], d2udt2[0] = _derivative_gaussian_integral(
                    ii, theta[0], theta[4], theta[2], PSFy
                )
                dudt[1], d2udt2[1] = _derivative_gaussian_integral(
                    jj, theta[1], theta[4], theta[2], PSFx
                )
                dudt[2] = PSFx * PSFy
                d2udt2[2] = 0.0
                dudt[3] = 1.0
                d2udt2[3] = 0.0
                dudt[4], d2udt2[4] = _derivative_gaussian_integral_2d_sigma(
                    ii, jj, theta[0], theta[1], theta[4], theta[2], PSFx, PSFy
                )

                model = theta[2] * dudt[2] + theta[3]
                cf = df = 0.0
                data = spot[ii, jj]
                if model > 10e-3:
                    cf = data / model - 1
                    df = data / model**2
                cf = np.minimum(cf, 10e4)
                df = np.minimum(df, 10e4)

                for ll in range(n_params):
                    numerator[ll] += cf * dudt[ll]
                    denominator[ll] += cf * d2udt2[ll] - df * dudt[ll] ** 2

        # The update
        for ll in range(n_params):
            if denominator[ll] == 0.0:
                update = np.sign(numerator[ll] * max_step[ll])
            else:
                update = np.minimum(
                    np.maximum(numerator[ll] / denominator[ll], -max_step[ll]),
                    max_step[ll],
                )
            if kk < 5:
                update *= GAMMA[ll]
            theta[ll] -= update

        # Other constraints
        theta[2] = np.maximum(theta[2], 1.0)
        theta[3] = np.maximum(theta[3], 0.01)
        theta[4] = np.maximum(theta[4], 0.01)
        theta[4] = np.minimum(theta[4], size)

        # Check for convergence
        if (np.abs(old_x - theta[0]) < eps) and (
            np.abs(old_y - theta[1]) < eps
        ):
            break
        else:
            old_x = theta[0]
            old_y = theta[1]

    thetas[index, 0:5] = theta
    thetas[index, 5] = theta[4]
    iterations[index] = kk

    # Calculating the CRLB and LogLikelihood
    Div = 0.0
    M = np.zeros((n_params, n_params), dtype=np.float32)
    for ii in range(size):
        for jj in range(size):
            PSFx = _gaussian_integral(ii, theta[0], theta[4])
            PSFy = _gaussian_integral(jj, theta[1], theta[4])
            model = theta[3] + theta[2] * PSFx * PSFy

            # Calculating derivatives
            dudt[0], d2udt2[0] = _derivative_gaussian_integral(
                ii, theta[0], theta[4], theta[2], PSFy
            )
            dudt[1], d2udt2[1] = _derivative_gaussian_integral(
                jj, theta[1], theta[4], theta[2], PSFx
            )
            dudt[4], d2udt2[4] = _derivative_gaussian_integral_2d_sigma(
                ii, jj, theta[0], theta[1], theta[4], theta[2], PSFx, PSFy
            )
            dudt[2] = PSFx * PSFy
            dudt[3] = 1.0

            # Building the Fisher Information Matrix
            model = theta[3] + theta[2] * dudt[2]
            for kk in range(n_params):
                for ll in range(kk, n_params):
                    M[kk, ll] += dudt[ll] * dudt[kk] / model
                    M[ll, kk] = M[kk, ll]

            # LogLikelihood
            if model > 0:
                data = spot[ii, jj]
                if data > 0:
                    Div += (
                        data * np.log(model)
                        - model
                        - data * np.log(data)
                        + data
                    )
                else:
                    Div += -model

    likelihoods[index] = Div

    # Matrix inverse (CRLB=F^-1)
    Minv = np.linalg.pinv(M)
    CRLB = np.zeros(n_params, dtype=np.float32)
    for kk in range(n_params):
        CRLB[kk] = Minv[kk, kk]
    CRLBs[index, 0:5] = CRLB
    CRLBs[index, 5] = CRLB[4]


@numba.jit(nopython=True, nogil=True)
def _mlefit_sigmaxy(
    spots: np.ndarray,
    index: int,
    thetas: np.ndarray,
    CRLBs: np.ndarray,
    likelihoods: np.ndarray,
    iterations: np.ndarray,
    eps: float,
    max_it: int,
) -> None:
    """Fit a Gaussian to a single spot using Maximum Likelihood
    Estimation (MLE) with separate sigmas for x and y dimensions."""
    n_params = 6

    spot = spots[index]
    size, _ = spot.shape

    # Initial values
    # theta is [x, y, N, bg, Sx, Sy]
    theta = _initial_theta_sigmaxy(spot, size)
    max_step = np.zeros(n_params, dtype=np.float32)
    max_step[0:2] = theta[4]
    max_step[2:4] = 0.1 * theta[2:4]
    max_step[4:6] = 0.2 * theta[4:6]

    # Memory allocation
    # (we do that outside of the loops to avoid huge delays in threaded code):
    dudt = np.zeros(n_params, dtype=np.float32)
    d2udt2 = np.zeros(n_params, dtype=np.float32)
    numerator = np.zeros(n_params, dtype=np.float32)
    denominator = np.zeros(n_params, dtype=np.float32)

    old_x = theta[0]
    old_y = theta[1]
    old_sx = theta[4]
    old_sy = theta[5]

    kk = 0
    while (
        kk < max_it
    ):  # we do this instead of a for loop for the special case of max_it=0
        kk += 1

        numerator[:] = 0.0
        denominator[:] = 0.0

        for ii in range(size):
            for jj in range(size):
                PSFx = _gaussian_integral(ii, theta[0], theta[4])
                PSFy = _gaussian_integral(jj, theta[1], theta[5])
                # Derivatives
                dudt[0], d2udt2[0] = _derivative_gaussian_integral(
                    ii, theta[0], theta[4], theta[2], PSFy
                )
                dudt[1], d2udt2[1] = _derivative_gaussian_integral(
                    jj, theta[1], theta[5], theta[2], PSFx
                )
                dudt[2] = PSFx * PSFy
                d2udt2[2] = 0.0
                dudt[3] = 1.0
                d2udt2[3] = 0.0
                dudt[4], d2udt2[4] = _derivative_gaussian_integral_1d_sigma(
                    ii, theta[0], theta[4], theta[2], PSFy
                )
                dudt[5], d2udt2[5] = _derivative_gaussian_integral_1d_sigma(
                    jj, theta[1], theta[5], theta[2], PSFx
                )

                model = theta[2] * dudt[2] + theta[3]
                cf = df = 0.0
                data = spot[ii, jj]
                if model > 10e-3:
                    cf = data / model - 1
                    df = data / model**2
                cf = np.minimum(cf, 10e4)
                df = np.minimum(df, 10e4)

                for ll in range(n_params):
                    numerator[ll] += cf * dudt[ll]
                    denominator[ll] += cf * d2udt2[ll] - df * dudt[ll] ** 2

        # The update
        for ll in range(n_params):
            if denominator[ll] == 0.0:
                # This is case is not handled in Lidke's code
                # but it seems to be a problem here
                # (maybe due to many iterations)
                theta[ll] -= GAMMA[ll] * np.sign(numerator[ll]) * max_step[ll]
            else:
                theta[ll] -= GAMMA[ll] * np.minimum(
                    np.maximum(numerator[ll] / denominator[ll], -max_step[ll]),
                    max_step[ll],
                )

        # Other constraints
        theta[2] = np.maximum(theta[2], 1.0)
        theta[3] = np.maximum(theta[3], 0.01)
        theta[4] = np.maximum(theta[4], 0.01)
        theta[5] = np.maximum(theta[5], 0.01)

        # Check for convergence
        if np.abs(old_x - theta[0]) < eps:
            if np.abs(old_y - theta[1]) < eps:
                if np.abs(old_sx - theta[4]) < eps:
                    if np.abs(old_sy - theta[5]) < eps:
                        break
        old_x = theta[0]
        old_y = theta[1]
        old_sx = theta[4]
        old_sy = theta[5]
    thetas[index] = theta
    iterations[index] = kk

    # Calculating the CRLB and LogLikelihood
    Div = 0.0
    M = np.zeros((n_params, n_params), dtype=np.float32)
    for ii in range(size):
        for jj in range(size):
            PSFx = _gaussian_integral(ii, theta[0], theta[4])
            PSFy = _gaussian_integral(jj, theta[1], theta[5])
            model = theta[3] + theta[2] * PSFx * PSFy

            # Calculating derivatives
            dudt[0], d2udt2[0] = _derivative_gaussian_integral(
                ii, theta[0], theta[4], theta[2], PSFy
            )
            dudt[1], d2udt2[1] = _derivative_gaussian_integral(
                jj, theta[1], theta[5], theta[2], PSFx
            )
            dudt[4], d2udt2[4] = _derivative_gaussian_integral_1d_sigma(
                ii, theta[0], theta[4], theta[2], PSFy
            )
            dudt[5], d2udt2[5] = _derivative_gaussian_integral_1d_sigma(
                jj, theta[1], theta[5], theta[2], PSFx
            )
            dudt[2] = PSFx * PSFy
            dudt[3] = 1.0

            # Building the Fisher Information Matrix
            model = theta[3] + theta[2] * dudt[2]
            for kk in range(n_params):
                for ll in range(kk, n_params):
                    M[kk, ll] += dudt[ll] * dudt[kk] / model
                    M[ll, kk] = M[kk, ll]

            # LogLikelihood
            if model > 0:
                data = spot[ii, jj]
                if data > 0:
                    Div += (
                        data * np.log(model)
                        - model
                        - data * np.log(data)
                        + data
                    )
                else:
                    Div += -model

    likelihoods[index] = Div

    # Matrix inverse (CRLB=F^-1)
    Minv = np.linalg.pinv(M)
    CRLB = np.zeros(n_params, dtype=np.float32)
    for kk in range(n_params):
        CRLB[kk] = Minv[kk, kk]
    CRLBs[index] = CRLB


def locs_from_fits(
    identifications: pd.DataFrame,
    theta: np.ndarray,
    CRLBs: np.ndarray,
    log_likelihoods: np.ndarray,
    iterations: np.ndarray,
    box: int,
) -> pd.DataFrame:
    """Convert the results of Gaussian fits into a data frame array
    suitable for further analysis or visualization.

    Parameters
    ----------
    identifications : pd.DataFrame
        Data frame containing the identifications of the
        spots, which should include 'frame', 'x', 'y' and
        'net_gradient'.
    theta : np.ndarray
        The fitted parameters for each spot, shape (N, 6) or (N, 5)
        depending on the method used.
    CRLBs : np.ndarray
        The Cramer-Rao Lower Bounds for the fitted parameters, shape
        (N, 6) or (N, 5).
    likelihoods : np.ndarray
        The log-likelihoods for each fitted spot, shape (N,).
    iterations : np.ndarray
        The number of iterations taken to converge for each spot,
        shape (N,).
    box : int
        The size of the box used for fitting, which is used to
        calculate the offsets for the x and y coordinates.

    Returns
    -------
    locs : pd.DataFrame
        DataFrame containing the fitted parameters and additional
        information for each spot, including frame, x, y, photons,
        sigma_x, sigma_y, background, localization precision
        (lpx, lpy), ellipticity, net gradient and identification ID
        (if available).
    """
    box_offset = int(box / 2)
    x = theta[:, 1] + identifications["x"] - box_offset
    y = theta[:, 0] + identifications["y"] - box_offset
    with np.errstate(invalid="ignore"):
        lpy = np.sqrt(CRLBs[:, 0])
        lpx = np.sqrt(CRLBs[:, 1])
        a = np.maximum(theta[:, 4], theta[:, 5])
        b = np.minimum(theta[:, 4], theta[:, 5])
        ellipticity = (a - b) / a

    locs = pd.DataFrame(
        {
            "frame": identifications["frame"].to_numpy(dtype=np.uint32),
            "x": x.astype(np.float32),
            "y": y.astype(np.float32),
            "photons": theta[:, 2].astype(np.float32),
            "sx": theta[:, 4].astype(np.float32),
            "sy": theta[:, 5].astype(np.float32),
            "bg": theta[:, 3].astype(np.float32),
            "lpx": lpx.astype(np.float32),
            "lpy": lpy.astype(np.float32),
            "ellipticity": ellipticity.astype(np.float32),
            "net_gradient": identifications["net_gradient"].astype(np.float32),
            "log_likelihood": log_likelihoods.astype(np.float32),
            "iterations": iterations.astype(np.uint32),
        }
    )
    if "n_id" in identifications.columns:
        locs["n_id"] = identifications.n_id.astype(np.uint32)
        locs.sort_values(by=["n_id"], kind="quicksort", inplace=True)
    else:
        locs.sort_values(by=["frame"], kind="quicksort", inplace=True)
    return locs


def sigma_uncertainty(
    sigma: pd.Series | np.ndarray,
    sigma_orth: pd.Series | np.ndarray,
    photons: pd.Series | np.ndarray,
    bg: pd.Series | np.ndarray,
) -> np.ndarray:
    """Calculate standard error of fitted sigma based on the MLE 2D
    Gaussian/Poisson noise model (picasso.gaussmle).

    Based on the approximation by Rieger and Stallinga, ChemPhysChem,
    2014.

    TODO: this is likely slightly incorrect for astigmatic imaging
    since spherical covariance Gaussian is assumed in Mortensen et
    al., Nat Methods, 2010.

    Parameters
    ----------
    sigma : pd.Series | np.ndarray
        Fitted sigma values in camera pixels.
    sigma_orth : pd.Series | np.ndarray
        Fitted sigma values in the orthogonal direction in camera
        pixels.
    photons : pd.Series | np.ndarray
        Number of photons.
    bg : pd.Series | np.ndarray
        Background photons per pixel.

    Returns
    -------
    se_sigma : np.ndarray
        Standard error of fitted sigma values in camera pixels.
    """
    sa2 = sigma**2 + 1 / 12
    tau = (2 * np.pi * sa2 * bg) / (photons)
    delta_sigma_sq = (sigma**2 / (4 * photons)) * (
        1 + 8 * tau + np.sqrt((8 * tau) / (1 + 2 * tau))
    )
    return np.sqrt(delta_sigma_sq)
